# sage_setup: distribution = sagemath-flint
# Deprecated header file; use sage/libs/flint/mag.pxd instead
# See https://github.com/sagemath/sage/pull/36449

from sage.libs.flint.types cimport mag_t

from sage.libs.flint.mag cimport (
    MAG_BITS,
    mag_init,
    mag_clear,
    mag_init_set,
    mag_swap,
    mag_set,
    mag_zero,
    mag_one,
    mag_inf,
    mag_is_special,
    mag_is_zero,
    mag_is_inf,
    mag_is_finite,
    mag_equal,
    mag_cmp,
    mag_cmp_2exp_si,
    mag_min,
    mag_max,
    mag_print,
    mag_randtest,
    mag_randtest_special,
    mag_set_d,
    mag_set_ui,
    mag_set_fmpz,
    mag_set_d_2exp_fmpz,
    mag_set_fmpz_2exp_fmpz,
    mag_set_ui_2exp_si,
    mag_get_fmpq,
    mag_set_ui_lower,
    mag_set_fmpz_lower,
    mag_set_fmpz_2exp_fmpz_lower,
    mag_mul_2exp_si,
    mag_mul_2exp_fmpz,
    mag_mul,
    mag_mul_ui,
    mag_mul_fmpz,
    mag_add,
    mag_addmul,
    mag_add_2exp_fmpz,
    mag_div,
    mag_div_ui,
    mag_div_fmpz,
    mag_mul_lower,
    mag_mul_ui_lower,
    mag_mul_fmpz_lower,
    mag_add_lower,
    mag_sub_lower,
    mag_fast_init_set,
    mag_fast_zero,
    mag_fast_is_zero,
    mag_fast_mul,
    mag_fast_addmul,
    mag_fast_add_2exp_si,
    mag_fast_mul_2exp_si,
    mag_pow_ui,
    mag_pow_fmpz,
    mag_pow_ui_lower,
    mag_sqrt,
    mag_rsqrt,
    mag_hypot,
    mag_log1p,
    mag_log_ui,
    mag_exp,
    mag_expm1,
    mag_exp_tail,
    mag_binpow_uiui,
    mag_fac_ui,
    mag_rfac_ui,
    mag_bernoulli_div_fac_ui,
    mag_polylog_tail)



# sage_setup: distribution = sagemath-flint
# Deprecated header file; use sage/libs/flint/acb.pxd instead
# See https://github.com/sagemath/sage/pull/36449

from sage.libs.flint.types cimport acb_struct, acb_t, acb_ptr, acb_srcptr

from sage.libs.flint.acb cimport (
    acb_realref,
    acb_imagref,
    acb_init,
    acb_clear,
    _acb_vec_init,
    _acb_vec_clear,
    acb_is_zero,
    acb_is_one,
    acb_is_finite,
    acb_is_exact,
    acb_is_int,
    acb_zero,
    acb_one,
    acb_onei,
    acb_set,
    acb_set_ui,
    acb_set_si,
    acb_set_fmpz,
    acb_set_arb,
    acb_set_fmpq,
    acb_set_round,
    acb_set_round_fmpz,
    acb_set_round_arb,
    acb_swap,
    acb_print,
    acb_printd,
    acb_randtest,
    acb_randtest_special,
    acb_randtest_precise,
    acb_randtest_param,
    acb_equal,
    acb_eq,
    acb_ne,
    acb_overlaps,
    acb_union,
    acb_get_abs_ubound_arf,
    acb_get_abs_lbound_arf,
    acb_get_rad_ubound_arf,
    acb_get_mag,
    acb_get_mag_lower,
    acb_contains_fmpq,
    acb_contains_fmpz,
    acb_contains,
    acb_contains_zero,
    acb_contains_int,
    acb_rel_error_bits,
    acb_rel_accuracy_bits,
    acb_bits,
    acb_indeterminate,
    acb_trim,
    acb_is_real,
    acb_get_unique_fmpz,
    acb_arg,
    acb_abs,
    acb_neg,
    acb_conj,
    acb_add_ui,
    acb_add_fmpz,
    acb_add_arb,
    acb_add,
    acb_sub_ui,
    acb_sub_fmpz,
    acb_sub_arb,
    acb_sub,
    acb_mul_onei,
    acb_div_onei,
    acb_mul_ui,
    acb_mul_si,
    acb_mul_fmpz,
    acb_mul_arb,
    acb_mul,
    acb_mul_2exp_si,
    acb_mul_2exp_fmpz,
    acb_cube,
    acb_addmul,
    acb_addmul_ui,
    acb_addmul_si,
    acb_addmul_fmpz,
    acb_addmul_arb,
    acb_submul,
    acb_submul_ui,
    acb_submul_si,
    acb_submul_fmpz,
    acb_submul_arb,
    acb_inv,
    acb_div_ui,
    acb_div_si,
    acb_div_fmpz,
    acb_div,
    acb_const_pi,
    acb_sqrt,
    acb_rsqrt,
    acb_pow_fmpz,
    acb_pow_ui,
    acb_pow_si,
    acb_pow_arb,
    acb_pow,
    acb_exp,
    acb_exp_pi_i,
    acb_exp_invexp,
    acb_log,
    acb_log1p,
    acb_sin,
    acb_cos,
    acb_sin_cos,
    acb_tan,
    acb_cot,
    acb_sec,
    acb_csc,
    acb_sin_pi,
    acb_cos_pi,
    acb_sin_cos_pi,
    acb_tan_pi,
    acb_cot_pi,
    acb_asin,
    acb_acos,
    acb_atan,
    acb_asinh,
    acb_acosh,
    acb_atanh,
    acb_lambertw,
    acb_sinh,
    acb_cosh,
    acb_sinh_cosh,
    acb_tanh,
    acb_coth,
    acb_sech,
    acb_csch,
    acb_rising_ui,
    acb_rising,
    acb_gamma,
    acb_rgamma,
    acb_lgamma,
    acb_digamma,
    acb_log_sin_pi,
    acb_polygamma,
    acb_barnes_g,
    acb_log_barnes_g,
    acb_zeta,
    acb_hurwitz_zeta,
    acb_polylog,
    acb_polylog_si,
    acb_agm1,
    acb_agm1_cpx,
    _acb_vec_init,
    _acb_vec_sort_pretty,
    _acb_vec_clear)



# sage_setup: distribution = sagemath-gap
# sage.doctest: needs sage.modules
r"""
Finitely generated abelian groups with GAP.

This module provides a python wrapper for abelian groups in GAP.

EXAMPLES::

    sage: from sage.groups.abelian_gps.abelian_group_gap import AbelianGroupGap
    sage: AbelianGroupGap([3,5])
    Abelian group with gap, generator orders (3, 5)

For infinite abelian groups we use the GAP package ``Polycyclic``::

    sage: AbelianGroupGap([3,0])   # optional - gap_package_polycyclic
    Abelian group with gap, generator orders (3, 0)

AUTHORS:

- Simon Brandhorst (2018-01-17): initial version
"""

# ****************************************************************************
#       Copyright (C) 2018 Simon Brandhorst <sbrandhorst@web.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  http://www.gnu.org/licenses/
# ****************************************************************************

from sage.groups.libgap_wrapper import ParentLibGAP, ElementLibGAP
from sage.groups.libgap_mixin import GroupMixinLibGAP
from sage.groups.group import AbelianGroup as AbelianGroupBase
from sage.libs.gap.element import GapElement
from sage.libs.gap.libgap import libgap
from sage.misc.cachefunc import cached_method
from sage.rings.integer_ring import ZZ
from sage.structure.unique_representation import UniqueRepresentation
from sage.categories.groups import Groups


class AbelianGroupElement_gap(ElementLibGAP):
    r"""
    An element of an abelian group via libgap.

    EXAMPLES::

        sage: from sage.groups.abelian_gps.abelian_group_gap import AbelianGroupGap
        sage: G = AbelianGroupGap([3,6])
        sage: G.gens()
        (f1, f2)
    """
    def __init__(self, parent, x, check=True):
        """
        The Python constructor.

        See :class:`AbelianGroupElement_gap` for details.

        INPUT:

        - ``parent`` -- an instance of :class:`AbelianGroup_gap`
        - ``x`` -- an instance of :class:`sage.libs.gap.element.GapElement`
        - ``check`` -- boolean (default: ``True``); check
          if ``x`` is an element  of the group

        TESTS::

            sage: from sage.groups.abelian_gps.abelian_group_gap import AbelianGroupGap
            sage: G = AbelianGroupGap([3,6])
            sage: g = G.an_element()
            sage: TestSuite(g).run()
        """
        if check and x not in parent.gap():
            raise ValueError("%s is not in the group %s" % (x, parent))
        ElementLibGAP.__init__(self, parent, x)

    def __hash__(self):
        r"""
        Return the hash of this element.

        EXAMPLES::

            sage: from sage.groups.abelian_gps.abelian_group_gap import AbelianGroupGap
            sage: G = AbelianGroupGap([3,2,4])
            sage: g = G.an_element()
            sage: g.__hash__()    # random
            1693277541873681615
        """
        return hash(self.parent()) ^ hash(self.exponents())

    def __reduce__(self):
        r"""
        Implement pickling.

        OUTPUT: tuple ``f`` such that this element is ``f[0](*f[1])``

        EXAMPLES::

            sage: from sage.groups.abelian_gps.abelian_group_gap import AbelianGroupGap
            sage: G = AbelianGroupGap([3,2,4])
            sage: g = G.an_element()
            sage: g == loads(dumps(g))
            True
            sage: g.__reduce__()
            (Abelian group with gap, generator orders (3, 2, 4), ((1, 1, 1),))
        """
        return self.parent(), (self.exponents(),)

    def exponents(self):
        r"""
        Return the tuple of exponents of this element.

        OUTPUT: tuple of integers

        EXAMPLES::

            sage: from sage.groups.abelian_gps.abelian_group_gap import AbelianGroupGap
            sage: G = AbelianGroupGap([4,7,9])
            sage: gens = G.gens()
            sage: g = gens[0]^2 * gens[1]^4 * gens[2]^8
            sage: g.exponents()
            (2, 4, 8)
            sage: S = G.subgroup(G.gens()[:1])
            sage: s = S.gens()[0]
            sage: s
            f1
            sage: s.exponents()
            (1,)

        It can handle quite large groups too::

            sage: G = AbelianGroupGap([2^10, 5^10])
            sage: f1, f2 = G.gens()
            sage: g = f1^123*f2^789
            sage: g.exponents()
            (123, 789)

        .. WARNING::

            Crashes for very large groups.

        .. TODO::

            Make exponents work for very large groups.
            This could be done by using Pcgs in gap.
        """

        P = self.parent()
        # better than Factorization as this does not create the
        # whole group in memory.
        f = P.gap().EpimorphismFromFreeGroup()
        x = f.PreImagesRepresentative(self.gap())
        L = x.ExtRepOfObj().sage()
        Lgens = L[::2]
        Lexpo = L[1::2]
        exp = []
        orders = P.gens_orders()
        i = 0
        for k in range(len(P.gens())):
            if k + 1 not in Lgens:
                exp.append(0)
            else:
                i = Lgens.index(k + 1)
                exp.append(Lexpo[i] % orders[k])
        return tuple(exp)

    def order(self):
        r"""
        Return the order of this element.

        OUTPUT: integer or infinity

        EXAMPLES::

            sage: from sage.groups.abelian_gps.abelian_group_gap import AbelianGroupGap
            sage: G = AbelianGroupGap([4])
            sage: g = G.gens()[0]
            sage: g.order()
            4
            sage: G = AbelianGroupGap([0])          # optional - gap_package_polycyclic
            sage: g = G.gens()[0]                   # optional - gap_package_polycyclic
            sage: g.order()                         # optional - gap_package_polycyclic
            +Infinity
        """
        return self.gap().Order().sage()


class AbelianGroupElement_polycyclic(AbelianGroupElement_gap):
    r"""
    An element of an abelian group using the GAP package ``Polycyclic``.

    TESTS::

        sage: from sage.groups.abelian_gps.abelian_group_gap import AbelianGroupGap
        sage: G = AbelianGroupGap([4,7,0])          # optional - gap_package_polycyclic
        sage: TestSuite(G.an_element()).run()       # optional - gap_package_polycyclic
    """
    def exponents(self):
        r"""
        Return the tuple of exponents of ``self``.

        OUTPUT: tuple of integers

        EXAMPLES::

            sage: # optional - gap_package_polycyclic
            sage: from sage.groups.abelian_gps.abelian_group_gap import AbelianGroupGap
            sage: G = AbelianGroupGap([4,7,0])
            sage: gens = G.gens()
            sage: g = gens[0]^2 * gens[1]^4 * gens[2]^8
            sage: g.exponents()
            (2, 4, 8)

        Efficiently handles very large groups::

            sage: # optional - gap_package_polycyclic
            sage: G = AbelianGroupGap([2^30,5^30,0])
            sage: f1, f2, f3 = G.gens()
            sage: (f1^12345*f2^123456789).exponents()
            (12345, 123456789, 0)
        """
        return tuple(self.gap().Exponents().sage())


class AbelianGroup_gap(UniqueRepresentation, GroupMixinLibGAP, ParentLibGAP, AbelianGroupBase):
    r"""
    Finitely generated abelian groups implemented in GAP.

    Needs the gap package ``Polycyclic`` in case the group is infinite.

    INPUT:

    - ``G`` -- a GAP group
    - ``category`` -- a category
    - ``ambient`` -- (optional) an :class:`AbelianGroupGap`

    EXAMPLES::

        sage: from sage.groups.abelian_gps.abelian_group_gap import AbelianGroupGap
        sage: G = AbelianGroupGap([3, 2, 5])
        sage: G
        Abelian group with gap, generator orders (3, 2, 5)
    """
    def __init__(self, G, category, ambient=None):
        r"""
        Create an instance of this class.

        See :class:`AbelianGroup_gap` for details

        TESTS::

            sage: from sage.groups.abelian_gps.abelian_group_gap import AbelianGroupGap
            sage: G = AbelianGroupGap([3,2,5])
            sage: TestSuite(G).run()
        """
        AbelianGroupBase.__init__(self, category=category)
        ParentLibGAP.__init__(self, G, ambient=ambient)

    Element = AbelianGroupElement_gap

    def _coerce_map_from_(self, S):
        r"""
        Return whether ``S`` canonically coerces to ``self``.

        INPUT:

        - ``S`` -- anything

        OUTPUT: boolean

        EXAMPLES::

            sage: from sage.groups.abelian_gps.abelian_group_gap import AbelianGroupGap
            sage: G = AbelianGroupGap([2,3,4,5])
            sage: gen = G.gens()[:2]
            sage: S = G.subgroup(gen)
            sage: G._coerce_map_from_(S)
            True
            sage: S._coerce_map_from_(G)
            False
        """
        if isinstance(S, AbelianGroup_gap):
            return S.is_subgroup_of(self)
        return super()._coerce_map_from_(S)

    def _element_constructor_(self, x, check=True):
        r"""
        Define coercions and conversions.

        INPUT:

        - ``x`` -- an element of this group, a GAP element

        EXAMPLES::

            sage: from sage.groups.abelian_gps.abelian_group_gap import AbelianGroupGap
            sage: G = AbelianGroupGap([2,3])
            sage: A = AbelianGroup([2,3])
            sage: a = A.an_element()
            sage: a
            f0*f1
            sage: G(a)
            f1*f2
            sage: A = AdditiveAbelianGroup([2,3])
            sage: a = A.an_element()
            sage: a
            (1, 0)
            sage: G(a)
            f1

        For general ``fgp_modules`` conversion is implemented if our
        group is in Smith form::

            sage: G = AbelianGroupGap([6])
            sage: A = ZZ^2
            sage: e0,e1 = A.gens()
            sage: A = A / A.submodule([2*e0, 3*e1])
            sage: a = 2 * A.an_element()
            sage: a
            (2)
            sage: G(a)
            f2

        TESTS:

        Document that :issue:`31428` is fixed::

            sage: A = AbelianGroupGap([])
            sage: A([]) == A.one()
            True
        """
        if isinstance(x, AbelianGroupElement_gap):
            try:
                if x in self._cover:
                    x = self._cover.gap().NaturalHomomorphismByNormalSubgroup(self._relations).Image(x.gap())
                else:
                    x = x.gap()
            except AttributeError:
                x = x.gap()
        elif x == 1 or x == ():
            x = self.gap().Identity()
        elif not isinstance(x, GapElement):
            from sage.groups.abelian_gps.abelian_group_element import AbelianGroupElement
            from sage.groups.additive_abelian.additive_abelian_group import AdditiveAbelianGroupElement
            from sage.modules.fg_pid.fgp_element import FGP_Element
            if isinstance(x, AbelianGroupElement):
                exp = x.exponents()
            elif isinstance(x, AdditiveAbelianGroupElement):
                exp = x._hermite_lift()
            elif isinstance(x, FGP_Element):
                exp = x.vector()
            else:
                from sage.modules.free_module_element import vector
                exp = vector(ZZ, x)
            # turn the exponents into a gap element
            gens_gap = self.gens()
            orders = self.gens_orders()
            if len(exp) != len(gens_gap):
                raise ValueError("input does not match the number of generators")
            x = self.one()
            for g, e, m in zip(gens_gap, exp, orders):
                if m != 0:
                    e = e % m
                x *= g**e
            x = x.gap()
        return self.element_class(self, x, check=check)

    def all_subgroups(self):
        r"""
        Return the list of all subgroups of this group.

        EXAMPLES::

            sage: from sage.groups.abelian_gps.abelian_group_gap import AbelianGroupGap
            sage: G = AbelianGroupGap([2, 3])
            sage: G.all_subgroups()
            [Subgroup of Abelian group with gap, generator orders (2, 3) generated by (),
             Subgroup of Abelian group with gap, generator orders (2, 3) generated by (f1,),
             Subgroup of Abelian group with gap, generator orders (2, 3) generated by (f2,),
             Subgroup of Abelian group with gap, generator orders (2, 3) generated by (f2, f1)]
        """
        subgroups_gap = self.gap().AllSubgroups()
        subgroups_sage = []
        for G in subgroups_gap:
            S = self.subgroup(G.GeneratorsOfGroup())
            subgroups_sage.append(S)
        return subgroups_sage

    def automorphism_group(self):
        r"""
        Return the group of automorphisms of ``self``.

        EXAMPLES::

            sage: from sage.groups.abelian_gps.abelian_group_gap import AbelianGroupGap
            sage: G = AbelianGroupGap([2, 3])
            sage: G.aut()
            Full group of automorphisms of Abelian group with gap, generator orders (2, 3)
        """
        from sage.groups.abelian_gps.abelian_aut import AbelianGroupAutomorphismGroup
        return AbelianGroupAutomorphismGroup(self)

    aut = automorphism_group

    def is_trivial(self):
        r"""
        Return ``True`` if this group is the trivial group.

        EXAMPLES::

            sage: from sage.groups.abelian_gps.abelian_group_gap import AbelianGroupGap
            sage: G = AbelianGroupGap([])
            sage: G
            Abelian group with gap, generator orders ()
            sage: G.is_trivial()
            True
            sage: AbelianGroupGap([1]).is_trivial()
            True
            sage: AbelianGroupGap([1,1,1]).is_trivial()
            True
            sage: AbelianGroupGap([2]).is_trivial()
            False
            sage: AbelianGroupGap([2,1]).is_trivial()
            False
        """
        return 1 == self.order()

    def identity(self):
        r"""
        Return the identity element of this group.

        EXAMPLES::

            sage: from sage.groups.abelian_gps.abelian_group_gap import AbelianGroupGap
            sage: G = AbelianGroupGap([4,10])
            sage: G.identity()
            1
        """
        return self.one()

    @cached_method
    def elementary_divisors(self):
        r"""
        Return the elementary divisors of this group.

        See :meth:`sage.groups.abelian_gps.abelian_group_gap.elementary_divisors`.

        EXAMPLES::

            sage: from sage.groups.abelian_gps.abelian_group_gap import AbelianGroupGap
            sage: G = AbelianGroupGap([2,3,4,5])
            sage: G.elementary_divisors()
            (2, 60)
        """
        ediv = self.gap().AbelianInvariants().sage()
        from sage.matrix.constructor import diagonal_matrix
        ed = diagonal_matrix(ZZ, ediv).elementary_divisors()
        return tuple(d for d in ed if d != 1)

    @cached_method
    def exponent(self):
        r"""
        Return the exponent of this abelian group.

        EXAMPLES::

            sage: from sage.groups.abelian_gps.abelian_group_gap import AbelianGroupGap
            sage: G = AbelianGroupGap([2,3,7])
            sage: G
            Abelian group with gap, generator orders (2, 3, 7)
            sage: G = AbelianGroupGap([2,4,6])
            sage: G
            Abelian group with gap, generator orders (2, 4, 6)
            sage: G.exponent()
            12
        """
        return self.gap().Exponent().sage()

    @cached_method
    def gens_orders(self):
        r"""
        Return the orders of the generators.

        Use :meth:`elementary_divisors` if you are looking for an
        invariant of the group.

        OUTPUT: tuple of integers

        EXAMPLES::

            sage: from sage.groups.abelian_gps.abelian_group_gap import AbelianGroupGap
            sage: Z2xZ3 = AbelianGroupGap([2,3])
            sage: Z2xZ3.gens_orders()
            (2, 3)
            sage: Z2xZ3.elementary_divisors()
            (6,)
            sage: Z6 = AbelianGroupGap([6])
            sage: Z6.gens_orders()
            (6,)
            sage: Z6.elementary_divisors()
            (6,)
            sage: Z2xZ3.is_isomorphic(Z6)
            True
            sage: Z2xZ3 is Z6
            False
        """
        from sage.rings.infinity import Infinity
        orders = []
        for g in self.gens():
            order = g.order()
            if order == Infinity:
                order = 0
            orders.append(order)
        return tuple(orders)

    def is_subgroup_of(self, G):
        r"""
        Return if ``self`` is a subgroup of ``G`` considered in
        the same ambient group.

        EXAMPLES::

            sage: from sage.groups.abelian_gps.abelian_group_gap import AbelianGroupGap
            sage: G = AbelianGroupGap([2,3,4,5])
            sage: gen = G.gens()[:2]
            sage: S1 = G.subgroup(gen)
            sage: S1.is_subgroup_of(G)
            True
            sage: S2 = G.subgroup(G.gens()[1:])
            sage: S2.is_subgroup_of(S1)
            False
        """
        if not isinstance(G, AbelianGroup_gap):
            raise ValueError("input must be an instance of AbelianGroup_gap")
        if self.ambient() is not G.ambient():
            return False
        return G.gap().IsSubsemigroup(self).sage()

    def _subgroup_constructor(self, libgap_subgroup):
        r"""
        Create a subgroup from the input.

        See :class:`~sage.groups.libgap_wrapper`. Override this in derived
        classes.

        EXAMPLES::

            sage: from sage.groups.abelian_gps.abelian_group_gap import AbelianGroupGap
            sage: A = AbelianGroupGap([2,3,4,5])
            sage: S = A.subgroup(A.gens()[:1])
            sage: T = A._subgroup_constructor(S.gap())
            sage: S is T
            True
        """
        ambient = self.ambient()
        generators = libgap_subgroup.GeneratorsOfGroup()
        generators = tuple([ambient(g) for g in generators])
        return AbelianGroupSubgroup_gap(ambient, generators)

    def quotient(self, N):
        r"""
        Return the quotient of this group by the normal subgroup `N`.

        INPUT:

        - ``N`` -- a subgroup
        - ``check`` -- boolean (default: ``True``); check if `N` is normal

        EXAMPLES::

            sage: from sage.groups.abelian_gps.abelian_group_gap import AbelianGroupGap
            sage: A = AbelianGroupGap([2,3,4,5])
            sage: S = A.subgroup(A.gens()[:1])
            sage: A.quotient(S)
            Quotient abelian group with generator orders (1, 3, 4, 5)
        """
        if not isinstance(N, AbelianGroup_gap):
            raise TypeError("not an abelian group")
        if not N.is_subgroup_of(self):
            raise ValueError("not a subgroup")
        return AbelianGroupQuotient_gap(self, N)

    def subgroup(self, gens):
        r"""
        Return the subgroup of this group generated by ``gens``.

        INPUT:

        - ``gens`` -- list of elements coercible into this group

        OUTPUT: a subgroup

        EXAMPLES::

            sage: from sage.groups.abelian_gps.abelian_group_gap import AbelianGroupGap
            sage: G = AbelianGroupGap([2,3,4,5])
            sage: gen = G.gens()[:2]
            sage: S = G.subgroup(gen)
            sage: S
            Subgroup of Abelian group with gap, generator orders (2, 3, 4, 5)
             generated by (f1, f2)
            sage: g = G.an_element()
            sage: s = S.an_element()
            sage: g * s
            f2^2*f3*f5

            sage: # optional - gap_package_polycyclic
            sage: G = AbelianGroupGap([3,4,0,2])
            sage: gen = G.gens()[:2]
            sage: S = G.subgroup(gen)
            sage: g = G.an_element()
            sage: s = S.an_element()
            sage: g * s
            g1^2*g2^2*g3*g4

        TESTS::

            sage: h = G.gens()[3]
            sage: h in S
            False
        """
        gens = tuple([self(g) for g in gens])
        return AbelianGroupSubgroup_gap(self.ambient(), gens)


class AbelianGroupGap(AbelianGroup_gap):
    r"""
    Abelian groups implemented using GAP.

    INPUT:

    - ``generator_orders`` -- list of nonnegative integers where `0`
      gives a factor isomorphic to `\ZZ`

    OUTPUT: an abelian group

    EXAMPLES::

        sage: from sage.groups.abelian_gps.abelian_group_gap import AbelianGroupGap
        sage: AbelianGroupGap([3,6])
        Abelian group with gap, generator orders (3, 6)
        sage: AbelianGroupGap([3,6,5])
        Abelian group with gap, generator orders (3, 6, 5)
        sage: AbelianGroupGap([3,6,0])      # optional - gap_package_polycyclic
        Abelian group with gap, generator orders (3, 6, 0)

    .. WARNING::

        Needs the GAP package ``Polycyclic`` in case the group is infinite.
    """
    @staticmethod
    def __classcall_private__(cls, generator_orders):
        r"""
        Normalize input to ensure a unique representation.

        EXAMPLES::

            sage: from sage.groups.abelian_gps.abelian_group_gap import AbelianGroupGap
            sage: A1 = AbelianGroupGap((2,3,4))
            sage: A2 = AbelianGroupGap([4/2,3,4])
            sage: A1 is A2
            True
        """
        generator_orders = tuple([ZZ(e) for e in generator_orders])
        if any(e < 0 for e in generator_orders):
            return ValueError("generator orders must be nonnegative")
        return super().__classcall__(cls, generator_orders)

    def __init__(self, generator_orders):
        r"""
        Constructor.

        TESTS::

            sage: from sage.groups.abelian_gps.abelian_group_gap import AbelianGroupGap
            sage: A = AbelianGroup((2,3,4))
            sage: TestSuite(A).run()
        """
        category = Groups().Commutative()
        if 0 in generator_orders:
            if not libgap.LoadPackage("Polycyclic"):
                raise ImportError("unable to import polycyclic package")
            G = libgap.eval("AbelianPcpGroup(%s)" % list(generator_orders))
            category = category.Infinite()
            self.Element = AbelianGroupElement_polycyclic
        else:
            G = libgap.AbelianGroup(generator_orders)
            category = category.Finite().Enumerated()
        AbelianGroup_gap.__init__(self, G, category=category)

    def _latex_(self):
        r"""
        Return a latex representation of this group.

        EXAMPLES::

            sage: from sage.groups.abelian_gps.abelian_group_gap import AbelianGroupGap
            sage: G = AbelianGroupGap([2,6])
            sage: G._latex_()
            \text{\texttt{Abelian group with gap, generator orders }} \left(2, 6\right)
        """
        from sage.misc.latex import latex
        base = r"\text{\texttt{Abelian group with gap, generator orders }}"
        return base + latex(self.gens_orders())

    def _repr_(self):
        r"""
        Return a string representation of this group.

        EXAMPLES::

            sage: from sage.groups.abelian_gps.abelian_group_gap import AbelianGroupGap
            sage: G = AbelianGroupGap([2,6])
            sage: G._repr_()
            'Abelian group with gap, generator orders (2, 6)'
        """
        return "Abelian group with gap, generator orders " + str(self.gens_orders())

    def __reduce__(self):
        r"""
        Implement pickling.

        We have to work around the fact that gap does not provide pickling.

        EXAMPLES::

            sage: from sage.groups.abelian_gps.abelian_group_gap import AbelianGroupGap
            sage: G = AbelianGroupGap([3,2,5])
            sage: G == loads(dumps(G))
            True
            sage: G is loads(dumps(G))
            True
        """
        return AbelianGroupGap, (self.gens_orders(),)


class AbelianGroupSubgroup_gap(AbelianGroup_gap):
    r"""
    Subgroups of abelian groups with GAP.

    INPUT:

    - ``ambient`` -- the ambient group
    - ``gens`` -- generators of the subgroup

    .. NOTE::

        Do not construct this class directly. Instead use
        :meth:`~sage.groups.abelian_groups.AbelianGroupGap.subgroup`.

    EXAMPLES::

        sage: from sage.groups.abelian_gps.abelian_group_gap import AbelianGroupGap
        sage: G = AbelianGroupGap([2,3,4,5])
        sage: gen = G.gens()[:2]
        sage: S = G.subgroup(gen)
    """
    def __init__(self, ambient, gens):
        r"""
        Initialize this subgroup.

        TESTS::

            sage: from sage.groups.abelian_gps.abelian_group_gap import AbelianGroupGap, AbelianGroupSubgroup_gap
            sage: G = AbelianGroupGap([])
            sage: gen = G.gens()
            sage: A = AbelianGroupSubgroup_gap(G, gen)
            sage: TestSuite(A).run()

        Check that we are in the correct category::

            sage: # optional - gap_package_polycyclic
            sage: G = AbelianGroupGap([2,3,0])
            sage: g = G.gens()
            sage: H1 = G.subgroup([g[0],g[1]])
            sage: H1 in Groups().Finite()
            True
            sage: H2 = G.subgroup([g[0],g[2]])
            sage: H2 in Groups().Infinite()
            True
        """
        gens_gap = tuple([g.gap() for g in gens])
        G = ambient.gap().Subgroup(gens_gap)
        from sage.rings.infinity import Infinity
        category = Groups().Commutative()
        if G.Size().sage() < Infinity:
            category = category.Finite()
        else:
            category = category.Infinite()
        category = category.Subobjects()
        AbelianGroup_gap.__init__(self, G, ambient=ambient, category=category)

    def _repr_(self):
        r"""
        Return a string representation of this subgroup.

        EXAMPLES::

            sage: from sage.groups.abelian_gps.abelian_group_gap import AbelianGroupGap
            sage: G = AbelianGroupGap([2,3,4,5])
            sage: gen = G.gens()[:2]
            sage: S = G.subgroup(gen)
            sage: S
            Subgroup of Abelian group with gap, generator orders (2, 3, 4, 5)
             generated by (f1, f2)
        """
        return "Subgroup of %s generated by %s" % (self.ambient(),self.gens())

    def __reduce__(self):
        r"""
        Implement pickling.

        We have to work around the fact that gap does not provide pickling.

        EXAMPLES::

            sage: from sage.groups.abelian_gps.abelian_group_gap import AbelianGroupGap
            sage: G = AbelianGroupGap([2,3,4,5])
            sage: gen = G.gens()[:2]
            sage: S = G.subgroup(gen)
            sage: S == loads(dumps(S))
            True
            sage: S is loads(dumps(S))
            True
        """
        amb = self.ambient()
        # avoid infinite loop
        gens = tuple([amb(g) for g in self.gens()])
        return amb.subgroup, (gens,)

    def lift(self, x):
        """
        Coerce to the ambient group.

        The terminology comes from the category framework and the more general notion of a subquotient.

        INPUT:

        - ``x`` -- an element of this subgroup

        OUTPUT: the corresponding element of the ambient group

        EXAMPLES::

            sage: from sage.groups.abelian_gps.abelian_group_gap import AbelianGroupGap
            sage: G = AbelianGroupGap([4])
            sage: g = G.gen(0)
            sage: H = G.subgroup([g^2])
            sage: h = H.gen(0); h
            f2
            sage: h.parent()
            Subgroup of Abelian group with gap, generator orders (4,) generated by (f2,)
            sage: H.lift(h)
            f2
            sage: H.lift(h).parent()
            Abelian group with gap, generator orders (4,)
        """
        return self.ambient()(x)

    def retract(self, x):
        """
        Convert an element of the ambient group into this subgroup.

        The terminology comes from the category framework and the more general
        notion of a subquotient.

        INPUT:

        - ``x`` -- an element of the ambient group that actually lies in this subgroup

        OUTPUT: the corresponding element of this subgroup

        EXAMPLES::

            sage: from sage.groups.abelian_gps.abelian_group_gap import AbelianGroupGap
            sage: G = AbelianGroupGap([4])
            sage: g = G.gen(0)
            sage: H = G.subgroup([g^2])
            sage: H.retract(g^2)
            f2
            sage: H.retract(g^2).parent()
            Subgroup of Abelian group with gap, generator orders (4,) generated by (f2,)
        """
        return self(x)


class AbelianGroupQuotient_gap(AbelianGroup_gap):
    r"""
    Quotients of abelian groups by a subgroup.

    .. NOTE::

        Do not call this directly. Instead use :meth:`quotient`.

    EXAMPLES::

        sage: from sage.groups.abelian_gps.abelian_group_gap import AbelianGroupGap
        sage: A = AbelianGroupGap([4,3])
        sage: N = A.subgroup([A.gen(0)^2])
        sage: Q1 = A.quotient(N)
        sage: Q1
        Quotient abelian group with generator orders (2, 3)
        sage: Q2 = Q1.quotient(Q1.subgroup(Q1.gens()[:1]))
        sage: Q2
        Quotient abelian group with generator orders (1, 3)
    """
    def __init__(self, G, N):
        r"""
        Constructor.

        TESTS::

            sage: from sage.groups.abelian_gps.abelian_group_gap import AbelianGroupGap
            sage: G = AbelianGroupGap([2,3,4,5])
            sage: gen = G.gens()[:2]
            sage: S = G.subgroup(gen)
            sage: Q = G.quotient(S)
            sage: TestSuite(Q).run()
        """
        self._cover = G
        self._relations = N
        libgap_group = G.gap().FactorGroup(N.gap())
        category = G.category()
        AbelianGroup_gap.__init__(self, libgap_group, category=category, ambient=None)

    def _repr_(self):
        r"""
        Return a string representation of this subgroup.

        EXAMPLES::

            sage: from sage.groups.abelian_gps.abelian_group_gap import AbelianGroupGap
            sage: G = AbelianGroupGap([2,3,4,5])
            sage: gen = G.gens()[:2]
            sage: S = G.subgroup(gen)
            sage: G.quotient(S)
            Quotient abelian group with generator orders (1, 1, 4, 5)
        """
        return "Quotient abelian group with generator orders " + str(
               self.gens_orders())

    def __reduce__(self):
        r"""
        Implement pickling.

        We have to work around the fact that gap does not provide pickling.

        EXAMPLES::

            sage: from sage.groups.abelian_gps.abelian_group_gap import AbelianGroupGap
            sage: A = AbelianGroupGap([4])
            sage: N = A.subgroup([A.gen(0)^2])
            sage: Q = A.quotient(N)
            sage: Q is loads(dumps(Q))
            True
        """
        G = self._cover
        N = self._relations
        return G.quotient, (N, )

    def _coerce_map_from_(self, S):
        r"""
        Return whether ``S`` canonically coerces to ``self``.

        INPUT:

        - ``S`` -- anything

        OUTPUT: boolean

        EXAMPLES::

            sage: from sage.groups.abelian_gps.abelian_group_gap import AbelianGroupGap
            sage: G = AbelianGroupGap([2,3,4,5])
            sage: gen = G.gens()[:2]
            sage: S = G.subgroup(gen)
            sage: Q = G.quotient(S)
            sage: Q._coerce_map_from_(G)
            True
            sage: G._coerce_map_from_(Q)
            False
        """
        if isinstance(S, AbelianGroup_gap):
            return self._cover._coerce_map_from_(S)

    def cover(self):
        r"""
        Return the covering group of this quotient group.

        EXAMPLES::

            sage: from sage.groups.abelian_gps.abelian_group_gap import AbelianGroupGap
            sage: G = AbelianGroupGap([2,3,4,5])
            sage: gen = G.gens()[:2]
            sage: S = G.subgroup(gen)
            sage: Q = G.quotient(S)
            sage: Q.cover() is G
            True
        """
        return self._cover

    def relations(self):
        r"""
        Return the relations of this quotient group.

        EXAMPLES::

            sage: from sage.groups.abelian_gps.abelian_group_gap import AbelianGroupGap
            sage: G = AbelianGroupGap([2,3,4,5])
            sage: gen = G.gens()[:2]
            sage: S = G.subgroup(gen)
            sage: Q = G.quotient(S)
            sage: Q.relations() is S
            True
        """
        return self._relations

    @cached_method
    def natural_homomorphism(self):
        r"""
        Return the defining homomorphism into ``self``.

        EXAMPLES::

            sage: from sage.groups.abelian_gps.abelian_group_gap import AbelianGroupGap
            sage: A = AbelianGroupGap([4])
            sage: N = A.subgroup([A.gen(0)^2])
            sage: Q = A.quotient(N)
            sage: Q.natural_homomorphism()
            Group morphism:
            From: Abelian group with gap, generator orders (4,)
            To:   Quotient abelian group with generator orders (2,)
        """
        phi = self._cover.gap().NaturalHomomorphismByNormalSubgroup(self._relations)
        Hom = self._cover.Hom(self)
        return Hom(phi)

    def lift(self, x):
        r"""
        Lift an element to the cover.

        EXAMPLES::

            sage: from sage.groups.abelian_gps.abelian_group_gap import AbelianGroupGap
            sage: A = AbelianGroupGap([4])
            sage: N = A.subgroup([A.gen(0)^2])
            sage: Q = A.quotient(N)
            sage: Q.lift(Q.0)
            f1
        """
        x = self(x)
        return self.natural_homomorphism().lift(x)



# sage_setup: distribution = sagemath-schemes
r"""
Composite morphisms of elliptic curves

It is often computationally convenient (for example, in cryptography)
to factor an isogeny of large degree into a composition of isogenies
of smaller (prime) degree. This class implements such a decomposition
while exposing (close to) the same interface as "normal", unfactored
elliptic-curve isogenies.

EXAMPLES:

The following example would take quite literally forever with the
straightforward :class:`EllipticCurveIsogeny` implementation, but
decomposing into prime steps is exponentially faster::

    sage: # needs sage.rings.finite_rings
    sage: p = 3 * 2^143 - 1
    sage: GF(p^2).inject_variables()
    Defining z2
    sage: E = EllipticCurve(GF(p^2), [1,0])
    sage: P = E.lift_x(31415926535897932384626433832795028841971 - z2)
    sage: P.order().factor()
    2^143
    sage: E.isogeny(P, algorithm="factored")  # preferred API
    Composite morphism of degree 11150372599265311570767859136324180752990208 = 2^143:
      From: Elliptic Curve defined by y^2 = x^3 + x
            over Finite Field in z2 of size 33451117797795934712303577408972542258970623^2
      To:   Elliptic Curve defined by y^2 = x^3 + (18676616716352953484576727486205473216172067*z2+32690199585974925193292786311814241821808308)*x
            + (3369702436351367403910078877591946300201903*z2+15227558615699041241851978605002704626689722)
            over Finite Field in z2 of size 33451117797795934712303577408972542258970623^2
    sage: from sage.schemes.elliptic_curves.hom_composite import EllipticCurveHom_composite
    sage: EllipticCurveHom_composite(E, P)  # same as above but use internal API, construct the isogeny directly
    Composite morphism of degree 11150372599265311570767859136324180752990208 = 2^143: ...

Yet, the interface provided by :class:`EllipticCurveHom_composite`
is identical to :class:`EllipticCurveIsogeny` and other instantiations
of :class:`EllipticCurveHom`::

    sage: # needs sage.rings.finite_rings
    sage: E = EllipticCurve(GF(419), [0,1])
    sage: P = E.lift_x(33); P.order()
    35
    sage: psi = EllipticCurveHom_composite(E, P); psi
    Composite morphism of degree 35 = 5*7:
      From: Elliptic Curve defined by y^2 = x^3 + 1 over Finite Field of size 419
      To:   Elliptic Curve defined by y^2 = x^3 + 101*x + 285 over Finite Field of size 419
    sage: psi(E.lift_x(11))
    (352 : 346 : 1)
    sage: psi.rational_maps()
    ((x^35 + 162*x^34 + 186*x^33 + 92*x^32 - ... + 44*x^3 + 190*x^2 + 80*x
      - 72)/(x^34 + 162*x^33 - 129*x^32 + 41*x^31 + ... + 66*x^3 - 191*x^2 + 119*x + 21),
     (x^51*y - 176*x^50*y + 115*x^49*y - 120*x^48*y + ... + 72*x^3*y + 129*x^2*y + 163*x*y
      + 178*y)/(x^51 - 176*x^50 + 11*x^49 + 26*x^48 - ... - 77*x^3 + 185*x^2 + 169*x - 128))
    sage: psi.kernel_polynomial()
    x^17 + 81*x^16 + 7*x^15 + 82*x^14 + 49*x^13 + 68*x^12 + 109*x^11 + 326*x^10
     + 117*x^9 + 136*x^8 + 111*x^7 + 292*x^6 + 55*x^5 + 389*x^4 + 175*x^3 + 43*x^2 + 149*x + 373
    sage: psi.dual()
    Composite morphism of degree 35 = 7*5:
      From: Elliptic Curve defined by y^2 = x^3 + 101*x + 285 over Finite Field of size 419
      To:   Elliptic Curve defined by y^2 = x^3 + 1 over Finite Field of size 419
    sage: psi.formal()
    t + 211*t^5 + 417*t^7 + 159*t^9 + 360*t^11 + 259*t^13 + 224*t^15 + 296*t^17 + 139*t^19 + 222*t^21 + O(t^23)

Equality is decided correctly (and, in some cases, much faster than
comparing :meth:`EllipticCurveHom.rational_maps`) even when distinct
factorizations of the same isogeny are compared::

    sage: psi == EllipticCurveIsogeny(E, P)                                             # needs sage.rings.finite_rings sage.symbolic
    True

We can easily obtain the individual factors of the composite map::

    sage: psi.factors()                                                                 # needs sage.rings.finite_rings
    (Isogeny of degree 5
      from Elliptic Curve defined by y^2 = x^3 + 1 over Finite Field of size 419
        to Elliptic Curve defined by y^2 = x^3 + 140*x + 214 over Finite Field of size 419,
     Isogeny of degree 7
      from Elliptic Curve defined by y^2 = x^3 + 140*x + 214 over Finite Field of size 419
        to Elliptic Curve defined by y^2 = x^3 + 101*x + 285 over Finite Field of size 419)

AUTHORS:

- Lukas Zobernig (2020): initial proof-of-concept version
- Lorenz Panny (2021): :class:`EllipticCurveHom` interface,
  documentation and tests, equality testing
"""

from sage.structure.richcmp import op_EQ
from sage.misc.cachefunc import cached_method
from sage.structure.sequence import Sequence

from sage.arith.misc import prod

from sage.schemes.elliptic_curves.ell_generic import EllipticCurve_generic
from sage.schemes.elliptic_curves.hom import EllipticCurveHom, compare_via_evaluation
from sage.schemes.elliptic_curves.ell_curve_isogeny import EllipticCurveIsogeny
from sage.schemes.elliptic_curves.weierstrass_morphism import WeierstrassIsomorphism, identity_morphism


def _eval_factored_isogeny(phis, P):
    """
    This method pushes a point `P` through a given sequence ``phis``
    of compatible isogenies.

    EXAMPLES::

        sage: # needs sage.rings.finite_rings
        sage: from sage.schemes.elliptic_curves import hom_composite
        sage: E = EllipticCurve(GF(419), [1,0])
        sage: Q = E(21, 8)
        sage: phis = []
        sage: while len(phis) < 10:
        ....:     P = list(sorted(E(0).division_points(7)))[1]
        ....:     phis.append(E.isogeny(P))
        ....:     E = phis[-1].codomain()
        sage: R = hom_composite._eval_factored_isogeny(phis, Q); R
        (290 : 183 : 1)
        sage: R in E
        True
    """
    for phi in phis:
        P = phi(P)
    return P


def _compute_factored_isogeny_prime_power(P, l, n, split=.8, velu_sqrt_bound=None):
    r"""
    This method takes a point `P` of order `\ell^n` and returns
    a sequence of degree-`\ell` isogenies whose composition has
    the subgroup generated by `P` as its kernel.

    The optional argument ``split``, a real number between
    `0` and `1`, controls the *strategy* used to compute the
    isogeny: In general, the algorithm performs a number of
    scalar multiplications `[\ell]` and a number of
    `\ell`-isogeny evaluations, and there exist tradeoffs
    between them.

    - Setting ``split`` to `0` skews the algorithm towards
      isogenies, minimizing multiplications.
      The asymptotic complexity is `O(n \log(\ell) + n^2 \ell)`.

    - Setting ``split`` to `1` skews the algorithm towards
      multiplications, minimizing isogenies.
      The asymptotic complexity is `O(n^2 \log(\ell) + n \ell)`.

    - Values strictly between `0` and `1` define *sparse*
      strategies, which balance the number of isogenies and
      multiplications according to the parameter.
      The asymptotic complexity is `O(n \log(n) \ell)`.

    The optional parameter ``velu_sqrt_bound`` prescribes the
    point in which the computation of a single isogeny should be
    performed using square root Velu instead of simple Velu.

    .. NOTE::

        As of July 2022, good values for ``split`` range somewhere
        between roughly `0.6` and `0.9`, depending on the size of
        `\ell` and the cost of base-field arithmetic.

    REFERENCES:

    Sparse strategies were introduced in [DJP2014]_, §4.2.2.

    ALGORITHM:

    The splitting rule implemented here is a simple heuristic which
    is usually not optimal. The advantage is that it does not rely
    on prior knowledge of degrees and exact costs of elliptic-curve
    arithmetic, while still working reasonably well for a fairly
    wide range of situations.

    EXAMPLES::

        sage: # needs sage.rings.finite_rings
        sage: from sage.schemes.elliptic_curves import hom_composite
        sage: E = EllipticCurve(GF(8191), [1,0])
        sage: P = E.random_point()
        sage: (l,n), = P.order().factor()
        sage: phis = hom_composite._compute_factored_isogeny_prime_power(P, l, n)
        sage: hom_composite._eval_factored_isogeny(phis, P)
        (0 : 1 : 0)
        sage: [phi.degree() for phi in phis] == [l]*n
        True

    All choices of ``split`` produce the same result, albeit
    not equally fast::

        sage: # needs sage.rings.finite_rings, long time (:issue:`39569`)
        sage: E = EllipticCurve(GF(2^127 - 1), [1,0])
        sage: P, = E.gens()
        sage: (l,n), = P.order().factor()
        sage: phis = hom_composite._compute_factored_isogeny_prime_power(P,l,n)
        sage: phis == hom_composite._compute_factored_isogeny_prime_power(P,l,n, split=0)  # long time -- about 10s
        True
        sage: phis == hom_composite._compute_factored_isogeny_prime_power(P,l,n, split=0.1)
        True
        sage: phis == hom_composite._compute_factored_isogeny_prime_power(P,l,n, split=0.5)
        True
        sage: phis == hom_composite._compute_factored_isogeny_prime_power(P,l,n, split=0.9)
        True
        sage: phis == hom_composite._compute_factored_isogeny_prime_power(P,l,n, split=1)
        True
    """
    def rec(Q, k):

        if k == 1:
            # base case: Q has order l
            Q._order = l # This was not cached before
            return [Q.curve().isogeny(kernel=Q, velu_sqrt_bound=velu_sqrt_bound)]

        # recursive case: k > 1 and Q has order l^k

        k1 = int(k * split + .5)
        k1 = max(1, min(k - 1, k1))  # clamp to [1, k - 1]

        Q1 = l**k1 * Q
        L = rec(Q1, k - k1)

        Q2 = _eval_factored_isogeny(L, Q)
        R = rec(Q2, k1)

        return L + R

    return rec(P, n)


def _compute_factored_isogeny_single_generator(P, velu_sqrt_bound=None):
    """
    This method takes a point `P` and returns a sequence of
    prime-degree isogenies whose composition has the subgroup
    generated by `P` as its kernel.

    The optional parameter ``velu_sqrt_bound`` prescribes the
    point in which the computation of a single isogeny should be
    performed using square root Velu instead of simple Velu.

    EXAMPLES::

        sage: # needs sage.rings.finite_rings
        sage: from sage.schemes.elliptic_curves import hom_composite
        sage: E = EllipticCurve(GF(419), [1,0])
        sage: P = E(42, 321)
        sage: phis = hom_composite._compute_factored_isogeny_single_generator(P)
        sage: list(sorted(phi.degree() for phi in phis))
        [2, 2, 3, 5, 7]
        sage: hom_composite._eval_factored_isogeny(phis, P)
        (0 : 1 : 0)

        ::

        sage: from sage.schemes.elliptic_curves import hom_composite
        sage: p = 3217
        sage: E = EllipticCurve_from_j(GF(p)(42))
        sage: P = E.gens()[0]
        sage: phis = hom_composite._compute_factored_isogeny_single_generator(P, velu_sqrt_bound=50)
        sage: for phi in phis:
        ....:     print(phi)
        Isogeny of degree 31 from Elliptic Curve defined by y^2 = x^3 + 114*x + 544 over Finite Field of size 3217 to Elliptic Curve defined by y^2 = x^3 + 277*x + 1710 over Finite Field of size 3217
        Elliptic-curve isogeny (using square-root Vélu) of degree 103:
          From: Elliptic Curve defined by y^2 = x^3 + 277*x + 1710 over Finite Field of size 3217
          To:   Elliptic Curve defined by y^2 = x^3 + 2979*x + 1951 over Finite Field of size 3217
    """
    phis = []
    h = P.order()
    for l,e in P.order().factor():
        h //= l**e
        psis = _compute_factored_isogeny_prime_power(h*P, l, e, velu_sqrt_bound=velu_sqrt_bound)
        P = _eval_factored_isogeny(psis, P)
        phis += psis
    return phis


def _compute_factored_isogeny(kernel, velu_sqrt_bound=None):
    """
    This method takes a set of points on an elliptic curve
    and returns a sequence of isogenies whose composition
    has the subgroup generated by that subset as its kernel.

    The optional parameter ``velu_sqrt_bound`` prescribes the
    point in which the computation of a single isogeny should be
    performed using square root Velu instead of simple Velu.

    EXAMPLES::

        sage: # needs sage.rings.finite_rings
        sage: from sage.schemes.elliptic_curves import hom_composite
        sage: E = EllipticCurve(GF(419), [-1,0])
        sage: Ps = [E(41,99), E(41,-99), E(51,14), E(21,21), E(33,17)]
        sage: phis = hom_composite._compute_factored_isogeny(Ps)
        sage: [phi.degree() for phi in phis]
        [2, 3, 5, 7, 2]
        sage: {hom_composite._eval_factored_isogeny(phis, P) for P in Ps}
        {(0 : 1 : 0)}
    """
    phis = []
    ker = list(kernel)
    while ker:
        K = ker.pop(0)
        psis = _compute_factored_isogeny_single_generator(K, velu_sqrt_bound=velu_sqrt_bound)
        ker = [_eval_factored_isogeny(psis, P) for P in ker]
        phis += psis
    return phis


class EllipticCurveHom_composite(EllipticCurveHom):

    _degree = None
    _phis = None

    def __init__(self, E, kernel, codomain=None, model=None, velu_sqrt_bound=None):
        """
        Construct a composite isogeny with given kernel (and optionally,
        prescribed codomain curve). The isogeny is decomposed into steps
        of prime degree.

        The ``codomain`` and ``model`` parameters have the same meaning
        as for :class:`EllipticCurveIsogeny`.

        The optional parameter ``velu_sqrt_bound`` prescribes the point
        in which the computation of a single isogeny should be performed
        using square root Velu instead of simple Velu. If not provided,
        the system default is used (see
        :class:`EllipticCurve_field.isogeny` for a more detailed
        discussion.

        EXAMPLES::

            sage: from sage.schemes.elliptic_curves.hom_composite import EllipticCurveHom_composite
            sage: E = EllipticCurve(GF(419), [1,0])                                     # needs sage.rings.finite_rings
            sage: EllipticCurveHom_composite(E, E.lift_x(23))                           # needs sage.rings.finite_rings
            Composite morphism of degree 105 = 3*5*7:
              From: Elliptic Curve defined by y^2 = x^3 + x
                    over Finite Field of size 419
              To:   Elliptic Curve defined by y^2 = x^3 + 373*x + 126
                    over Finite Field of size 419

        The given kernel generators need not be independent::

            sage: # needs database_cremona_mini_ellcurve sage.rings.number_field
            sage: x = polygen(ZZ, 'x')
            sage: K.<a> = NumberField(x^2 - x - 5)
            sage: E = EllipticCurve('210.b6').change_ring(K)
            sage: E.torsion_subgroup()
            Torsion Subgroup isomorphic to Z/12 + Z/2 associated to the Elliptic Curve
             defined by y^2 + x*y + y = x^3 + (-578)*x + 2756
              over Number Field in a with defining polynomial x^2 - x - 5
            sage: EllipticCurveHom_composite(E, E.torsion_points())
            Composite morphism of degree 24 = 2^3*3:
              From: Elliptic Curve defined by y^2 + x*y + y = x^3 + (-578)*x + 2756
                    over Number Field in a with defining polynomial x^2 - x - 5
              To:   Elliptic Curve defined by
                    y^2 + x*y + y = x^3 + (-89915533/16)*x + (-328200928141/64)
                    over Number Field in a with defining polynomial x^2 - x - 5

        TESTS::

            sage: E = EllipticCurve(GF(19), [1,0])
            sage: P = E.random_point()
            sage: psi = EllipticCurveHom_composite(E, P)
            sage: psi   # random
            Composite morphism of degree 10 = 2*5:
              From: Elliptic Curve defined by y^2 = x^3 + x over Finite Field of size 19
              To:   Elliptic Curve defined by y^2 = x^3 + 14*x over Finite Field of size 19

        ::

            sage: EllipticCurveHom_composite(E, E.lift_x(3), codomain=E)
            Composite morphism of degree 20 = 2^2*5:
              From: Elliptic Curve defined by y^2 = x^3 + x over Finite Field of size 19
              To:   Elliptic Curve defined by y^2 = x^3 + x over Finite Field of size 19

        ::

            sage: # needs sage.rings.finite_rings
            sage: E = EllipticCurve(GF((2^127-1)^2), [1,0])
            sage: K = 2^30 * E.random_point()
            sage: psi = EllipticCurveHom_composite(E, K, model='montgomery')
            sage: psi.codomain().a_invariants()
            (0, ..., 0, 1, 0)
        """
        if not isinstance(E, EllipticCurve_generic):
            raise ValueError(f'not an elliptic curve: {E}')

        if not isinstance(kernel, list) and not isinstance(kernel, tuple):
            kernel = [kernel]

        for P in kernel:
            if P not in E:
                raise ValueError(f'given point {P} does not lie on {E}')

        self._phis = _compute_factored_isogeny(kernel, velu_sqrt_bound=velu_sqrt_bound)

        if not self._phis:
            self._phis = [identity_morphism(E)]

        if model is not None:
            if codomain is not None:
                raise ValueError("cannot specify a codomain curve and model name simultaneously")

            from sage.schemes.elliptic_curves.ell_field import compute_model
            codomain = compute_model(self._phis[-1].codomain(), model)

        if codomain is not None:
            if not isinstance(codomain, EllipticCurve_generic):
                raise ValueError(f'not an elliptic curve: {codomain}')
            iso = self._phis[-1].codomain().isomorphism_to(codomain)
            if hasattr(self._phis[-1], '_set_post_isomorphism'):
                self._phis[-1]._set_post_isomorphism(iso)
            else:
                self._phis.append(iso)

        self._phis = tuple(self._phis)  # make immutable
        self.__perform_inheritance_housekeeping()

    def __perform_inheritance_housekeeping(self):
        """
        Internal helper function to set values of the superclass.

        TESTS::

            sage: from sage.schemes.elliptic_curves.hom_composite import EllipticCurveHom_composite
            sage: E = EllipticCurve([1,0])
            sage: phi = EllipticCurveHom_composite(E, E(0,0))   # implicit doctest
            sage: from sage.schemes.elliptic_curves.hom import EllipticCurveHom
            sage: print(EllipticCurveHom._repr_(phi))
            Elliptic-curve morphism:
              From: Elliptic Curve defined by y^2 = x^3 + x over Rational Field
              To:   Elliptic Curve defined by y^2 = x^3 - 4*x over Rational Field
            sage: phi.domain()
            Elliptic Curve defined by y^2 = x^3 + x over Rational Field
            sage: phi.codomain()
            Elliptic Curve defined by y^2 = x^3 - 4*x over Rational Field
        """
        self._degree = prod(phi.degree() for phi in self._phis)
        self._domain = self._phis[0].domain()
        self._codomain = self._phis[-1].codomain()
        EllipticCurveHom.__init__(self, self._domain, self._codomain)

    @classmethod
    def from_factors(cls, maps, E=None, strict=True):
        r"""
        This method constructs a :class:`EllipticCurveHom_composite`
        object encapsulating a given sequence of compatible isogenies.

        The isogenies are composed in left-to-right order, i.e., the
        resulting composite map equals `f_{n-1} \circ \dots \circ f_0`
        where `f_i` denotes ``maps[i]``.

        INPUT:

        - ``maps`` -- sequence of :class:`EllipticCurveHom` objects
        - ``E`` -- (optional) the domain elliptic curve
        - ``strict`` -- boolean (default: ``True``); if ``True``,
          always return an :class:`EllipticCurveHom_composite` object,
          else may return another :class:`EllipticCurveHom` type

        OUTPUT: the composite of ``maps``

        EXAMPLES::

            sage: from sage.schemes.elliptic_curves.hom_composite import EllipticCurveHom_composite
            sage: E = EllipticCurve(GF(43), [1,0])
            sage: P, = E.gens()
            sage: phi = EllipticCurveHom_composite(E, P)
            sage: psi = EllipticCurveHom_composite.from_factors(phi.factors())
            sage: psi == phi                                                            # needs sage.symbolic
            True

        TESTS::

            sage: E = EllipticCurve('4730k1')
            sage: EllipticCurveHom_composite.from_factors([], E) == E.scalar_multiplication(1)
            True

        ::

            sage: # needs sage.rings.finite_rings
            sage: E = EllipticCurve(GF(419), [1,0])
            sage: P, = E.gens()
            sage: phi = EllipticCurveHom_composite(E, P)
            sage: EllipticCurveHom_composite.from_factors(phi.factors()) == phi         # needs sage.symbolic
            True
        """
        maps = tuple(maps)
        if not maps and E is None:
            raise ValueError('need either factors or domain')
        if E is None:
            E = maps[0].domain()

        for phi in maps:
            if not isinstance(phi, EllipticCurveHom):
                raise TypeError(f'not an elliptic-curve isogeny: {phi}')
            if phi.domain() != E:
                raise ValueError(f'isogeny has incorrect domain: {phi}')
            E = phi.codomain()

        if not maps:
            maps = (identity_morphism(E),)

        if len(maps) == 1 and not strict:
            return maps[0]

        result = cls.__new__(cls)
        result._phis = maps
        result.__perform_inheritance_housekeeping()
        return result

    def _call_(self, P):
        """
        Evaluate this composite isogeny at a point.

        TESTS::

            sage: # needs database_cremona_mini_ellcurve sage.rings.number_field
            sage: from sage.schemes.elliptic_curves.hom_composite import EllipticCurveHom_composite
            sage: x = polygen(ZZ, 'x')
            sage: K.<a> = NumberField(x^2 - x - 5)
            sage: E = EllipticCurve('210.b6').change_ring(K)
            sage: psi = EllipticCurveHom_composite(E, E.torsion_points())
            sage: R = E.lift_x(15/4 * (a+3))
            sage: psi(R)    # indirect doctest
            (1033648757/303450 : -58397496786187/1083316500*a + 54706287407197/2166633000 : 1)

        Check that copying the order over works::

            sage: # needs sage.rings.finite_rings
            sage: E = EllipticCurve(GF(431), [1,0])
            sage: P, = E.gens()
            sage: Q = 2^99*P; Q.order()
            27
            sage: phi = E.isogeny(3^99*P, algorithm='factored')
            sage: phi(Q)._order
            27
        """
        return _eval_factored_isogeny(self._phis, P)

    def _eval(self, P):
        r"""
        Less strict evaluation method for internal use.

        In particular, this can be used to evaluate ``self`` at a
        point defined over an extension field.

        INPUT: a sequence of 3 coordinates defining a point on ``self``

        OUTPUT: the result of evaluating ``self`` at the given point

        EXAMPLES::

            sage: from sage.schemes.elliptic_curves.hom_composite import EllipticCurveHom_composite
            sage: E = EllipticCurve(j=Mod(1728,419))
            sage: K, = E.gens()
            sage: psi = EllipticCurveHom_composite(E, 4*K)
            sage: Ps = E.change_ring(GF(419**2))(0).division_points(5)                              # needs sage.rings.finite_rings
            sage: {psi._eval(P).curve() for P in Ps}                                                # needs sage.rings.finite_rings
            {Elliptic Curve defined by y^2 = x^3 + 373*x + 126 over Finite Field in z2 of size 419^2}
        """
        if self._domain.defining_polynomial()(*P):
            raise ValueError(f'{P} not on {self._domain}')
        k = Sequence(P).universe()

        Q = P
        for phi in self._phis:
            Q = phi._eval(Q)

        return self._codomain.base_extend(k)(*Q)

    def _repr_(self):
        """
        Return basic facts about this composite isogeny as a string.

        TESTS::

            sage: from sage.schemes.elliptic_curves.hom_composite import EllipticCurveHom_composite
            sage: E = EllipticCurve(GF(43), [1,0])
            sage: P, = E.gens()
            sage: phi = EllipticCurveHom_composite(E, P)
            sage: phi   # indirect doctest
            Composite morphism of degree 44 = 2^2*11:
              From: Elliptic Curve defined by y^2 = x^3 + x over Finite Field of size 43
              To:   Elliptic Curve defined by y^2 = x^3 + x over Finite Field of size 43
            sage: phi * phi * phi * phi * phi * phi * phi   # indirect doctest
            Composite morphism of degree 319277809664 = 2^2*11*2^2*11*2^2*11*2^2*11*2^2*11*2^2*11*2^2*11:
              From: Elliptic Curve defined by y^2 = x^3 + x over Finite Field of size 43
              To:   Elliptic Curve defined by y^2 = x^3 + x over Finite Field of size 43
        """
        from itertools import groupby
        degs = [phi.degree() for phi in self._phis]
        if len(degs) == 1:
            return f'Composite morphism of degree {self._degree}:' \
                    f'\n  From: {self._domain}' \
                    f'\n  To:   {self._codomain}'
        grouped = [(d, sum(1 for _ in g)) for d,g in groupby(degs)]
        degs_str = '*'.join(str(d) + (f'^{e}' if e > 1 else '') for d,e in grouped)
        return f'Composite morphism of degree {self._degree} = {degs_str}:' \
                f'\n  From: {self._domain}' \
                f'\n  To:   {self._codomain}'

    def factors(self):
        r"""
        Return the factors of this composite isogeny as a tuple.

        The isogenies are returned in left-to-right order, i.e.,
        the returned tuple `(f_1,...,f_n)` corresponds to the map
        `f_n \circ \dots \circ f_1`.

        EXAMPLES::

            sage: from sage.schemes.elliptic_curves.hom_composite import EllipticCurveHom_composite
            sage: E = EllipticCurve(GF(43), [1,0])
            sage: P, = E.gens()
            sage: phi = EllipticCurveHom_composite(E, P)
            sage: phi.factors()
            (Isogeny of degree 2
              from Elliptic Curve defined by y^2 = x^3 + x over Finite Field of size 43
                to Elliptic Curve defined by y^2 = x^3 + 39*x over Finite Field of size 43,
             Isogeny of degree 2
              from Elliptic Curve defined by y^2 = x^3 + 39*x over Finite Field of size 43
                to Elliptic Curve defined by y^2 = x^3 + 42*x + 26 over Finite Field of size 43,
             Isogeny of degree 11
              from Elliptic Curve defined by y^2 = x^3 + 42*x + 26 over Finite Field of size 43
                to Elliptic Curve defined by y^2 = x^3 + x over Finite Field of size 43)
        """
        return self._phis

    # EllipticCurveHom methods

    @staticmethod
    def _composition_impl(left, right):
        """
        Helper method to compose other elliptic-curve morphisms with
        :class:`EllipticCurveHom_composite` objects. Called by
        :meth:`EllipticCurveHom._composition_`.

        TESTS::

            sage: # needs sage.rings.number_field
            sage: from sage.schemes.elliptic_curves.hom_composite import EllipticCurveHom_composite
            sage: E = EllipticCurve([i + 1, i, 0, -4, -6*i])
            sage: P,Q = E.lift_x(i - 5), E.lift_x(-4*i)
            sage: phi = EllipticCurveHom_composite(E, P)
            sage: psi = phi.codomain().isogeny(phi(Q))
            sage: from sage.schemes.elliptic_curves.weierstrass_morphism import WeierstrassIsomorphism
            sage: iso1 = WeierstrassIsomorphism(E, (-1, 0, -i - 1, 0))
            sage: iso2 = psi.codomain().isomorphism_to(E)
            sage: psi * phi                 # indirect doctest
            Composite morphism of degree 16 = 2^2*4:
              From: Elliptic Curve defined by y^2 + (I+1)*x*y = x^3 + I*x^2 + (-4)*x + (-6*I)
                    over Number Field in I with defining polynomial x^2 + 1 with I = 1*I
              To:   Elliptic Curve defined by y^2 + (I+1)*x*y = x^3 + I*x^2 + (-3331/4)*x + (-142593/8*I)
                    over Number Field in I with defining polynomial x^2 + 1 with I = 1*I
            sage: iso2 * EllipticCurveHom_composite.from_factors([phi, psi]) # indirect doctest
            Composite morphism of degree 16 = 4^2:
              From: Elliptic Curve defined by y^2 + (I+1)*x*y = x^3 + I*x^2 + (-4)*x + (-6*I)
                    over Number Field in I with defining polynomial x^2 + 1 with I = 1*I
              To:   Elliptic Curve defined by y^2 + (I+1)*x*y = x^3 + I*x^2 + (-4)*x + (-6*I)
                    over Number Field in I with defining polynomial x^2 + 1 with I = 1*I
            sage: phi * iso1                # indirect doctest
            Composite morphism of degree 4 = 2^2:
              From: Elliptic Curve defined by y^2 + (I+1)*x*y = x^3 + I*x^2 + (-4)*x + (-6*I)
                    over Number Field in I with defining polynomial x^2 + 1 with I = 1*I
              To:   Elliptic Curve defined by y^2 + (I+1)*x*y = x^3 + I*x^2 + (480*I-694)*x + (-7778*I+5556)
                    over Number Field in I with defining polynomial x^2 + 1 with I = 1*I
            sage: iso2 * psi * phi * iso1   # indirect doctest
            Composite morphism of degree 16 = 2^2*4:
              From: Elliptic Curve defined by y^2 + (I+1)*x*y = x^3 + I*x^2 + (-4)*x + (-6*I)
                    over Number Field in I with defining polynomial x^2 + 1 with I = 1*I
              To:   Elliptic Curve defined by y^2 + (I+1)*x*y = x^3 + I*x^2 + (-4)*x + (-6*I)
                    over Number Field in I with defining polynomial x^2 + 1 with I = 1*I
        """
        if isinstance(left, EllipticCurveHom_composite):
            if isinstance(right, WeierstrassIsomorphism) and hasattr(left.factors()[0], '_set_pre_isomorphism'):    # XXX bit of a hack
                return EllipticCurveHom_composite.from_factors((left.factors()[0] * right,) + left.factors()[1:], strict=False)
            if isinstance(right, EllipticCurveHom_composite):
                return EllipticCurveHom_composite.from_factors(right.factors() + left.factors())
            if isinstance(right, EllipticCurveHom):
                return EllipticCurveHom_composite.from_factors((right,) + left.factors())
        if isinstance(right, EllipticCurveHom_composite):
            if isinstance(left, WeierstrassIsomorphism) and hasattr(right.factors()[-1], '_set_post_isomorphism'):  # XXX bit of a hack
                return EllipticCurveHom_composite.from_factors(right.factors()[:-1] + (left * right.factors()[-1],), strict=False)
            if isinstance(left, EllipticCurveHom):
                return EllipticCurveHom_composite.from_factors(right.factors() + (left,))
        return NotImplemented

    @staticmethod
    def _comparison_impl(left, right, op):
        r"""
        Compare a composite isogeny to another elliptic-curve morphism.

        Called by :meth:`EllipticCurveHom._richcmp_`.

        ALGORITHM:

        If possible, we use
        :func:`~sage.schemes.elliptic_curves.hom.compare_via_evaluation`.
        The complexity in that case is polynomial in the representation
        size of this morphism.

        TESTS::

            sage: # needs sage.rings.number_field
            sage: from sage.schemes.elliptic_curves.hom_composite import EllipticCurveHom_composite
            sage: E = EllipticCurve(QuadraticField(-3), [0,16])
            sage: P,Q = E.lift_x(0), E.lift_x(-4)
            sage: phi = EllipticCurveHom_composite(E, P)
            sage: psi = phi.codomain().isogeny(phi(Q))
            sage: psi = psi.codomain().isomorphism_to(E) * psi
            sage: comp = psi * phi
            sage: mu = E.scalar_multiplication(phi.degree())
            sage: sum(a*comp == mu for a in E.automorphisms())
            1

        ::

            sage: # needs sage.rings.finite_rings sage.rings.number_field sage.symbolic
            sage: E = EllipticCurve(GF(431**2), [1,0])
            sage: P,Q = E.gens()
            sage: phi1 = EllipticCurveHom_composite(E, P)
            sage: phi2 = EllipticCurveHom_composite(phi1.codomain(), phi1(Q))
            sage: psi1 = EllipticCurveHom_composite(E, Q)
            sage: psi2 = EllipticCurveHom_composite(psi1.codomain(), psi1(P))
            sage: phi2 * phi1 == psi2 * psi1
            True
        """
        if op != op_EQ:
            return NotImplemented
        try:
            return compare_via_evaluation(left, right)
        except NotImplementedError:
            return NotImplemented

    def rational_maps(self):
        """
        Return the pair of explicit rational maps defining this composite
        isogeny.

        EXAMPLES::

            sage: # needs sage.rings.finite_rings
            sage: from sage.schemes.elliptic_curves.hom_composite import EllipticCurveHom_composite
            sage: E = EllipticCurve(GF(65537), [1,2,3,4,5])
            sage: P = E.lift_x(7321)
            sage: phi = EllipticCurveHom_composite(E, P)
            sage: phi.rational_maps()
            ((x^9 + 27463*x^8 + 21204*x^7 - 5750*x^6 + 1610*x^5 + 14440*x^4
              + 26605*x^3 - 15569*x^2 - 3341*x + 1267)/(x^8 + 27463*x^7 + 26871*x^6
              + 5999*x^5 - 20194*x^4 - 6310*x^3 + 24366*x^2 - 20905*x - 13867),
             (x^12*y + 8426*x^11*y + 5667*x^11 + 27612*x^10*y + 26124*x^10 + 9688*x^9*y
              - 22715*x^9 + 19864*x^8*y + 498*x^8 + 22466*x^7*y - 14036*x^7 + 8070*x^6*y
              + 19955*x^6 - 20765*x^5*y - 12481*x^5 + 12672*x^4*y + 24142*x^4 - 23695*x^3*y
              + 26667*x^3 + 23780*x^2*y + 17864*x^2 + 15053*x*y - 30118*x + 17539*y
              - 23609)/(x^12 + 8426*x^11 + 21945*x^10 - 22587*x^9 + 22094*x^8 + 14603*x^7
              - 26255*x^6 + 11171*x^5 - 16508*x^4 - 14435*x^3 - 2170*x^2 + 29081*x - 19009))

        TESTS::

            sage: f = phi.codomain().defining_polynomial()                              # needs sage.rings.finite_rings
            sage: g = E.defining_polynomial().subs({2:1})                               # needs sage.rings.finite_rings
            sage: f(*phi.rational_maps(), 1) % g                                        # needs sage.rings.finite_rings
            0

        ::

            sage: phi.rational_maps()[0].parent()                                       # needs sage.rings.finite_rings
            Fraction Field of
             Multivariate Polynomial Ring in x, y over Finite Field of size 65537
            sage: phi.rational_maps()[1].parent()                                       # needs sage.rings.finite_rings
            Fraction Field of
             Multivariate Polynomial Ring in x, y over Finite Field of size 65537
        """
        fx, fy = self._phis[-1].rational_maps()
        for phi in self._phis[:-1][::-1]:
            gx, gy = phi.rational_maps()
            fx, fy = fx(gx, gy), fy(gx, gy)
        return (fx, fy)

    def x_rational_map(self):
        """
        Return the `x`-coordinate rational map of this composite isogeny.

        EXAMPLES::

            sage: # needs sage.rings.finite_rings
            sage: from sage.schemes.elliptic_curves.hom_composite import EllipticCurveHom_composite
            sage: E = EllipticCurve(GF(65537), [1,2,3,4,5])
            sage: P = E.lift_x(7321)
            sage: phi = EllipticCurveHom_composite(E, P)
            sage: phi.x_rational_map() == phi.rational_maps()[0]
            True

        TESTS::

            sage: phi.x_rational_map().parent()                                         # needs sage.rings.finite_rings
            Fraction Field of Univariate Polynomial Ring in x
             over Finite Field of size 65537
        """
        fx = self._phis[-1].x_rational_map()
        for phi in self._phis[:-1][::-1]:
            fx = fx(phi.x_rational_map())
        return fx

    def kernel_polynomial(self):
        """
        Return the kernel polynomial of this composite isogeny.

        EXAMPLES::

            sage: # needs sage.rings.finite_rings
            sage: from sage.schemes.elliptic_curves.hom_composite import EllipticCurveHom_composite
            sage: E = EllipticCurve(GF(65537), [1,2,3,4,5])
            sage: P = E.lift_x(7321)
            sage: phi = EllipticCurveHom_composite(E, P); phi
            Composite morphism of degree 9 = 3^2:
              From: Elliptic Curve defined by y^2 + x*y + 3*y = x^3 + 2*x^2 + 4*x + 5
                    over Finite Field of size 65537
              To:   Elliptic Curve defined by y^2 + x*y + 3*y = x^3 + 2*x^2 + 28339*x + 59518
                    over Finite Field of size 65537
            sage: phi.kernel_polynomial()
            x^4 + 46500*x^3 + 19556*x^2 + 7643*x + 15952
        """
        # shouldn't there be a better algorithm for this?
        return self.x_rational_map().denominator().radical()

    @cached_method
    def dual(self):
        """
        Return the dual of this composite isogeny.

        EXAMPLES::

            sage: # needs sage.rings.finite_rings
            sage: from sage.schemes.elliptic_curves.hom_composite import EllipticCurveHom_composite
            sage: E = EllipticCurve(GF(65537), [1,2,3,4,5])
            sage: P = E.lift_x(7321)
            sage: phi = EllipticCurveHom_composite(E, P); phi
            Composite morphism of degree 9 = 3^2:
              From: Elliptic Curve defined by y^2 + x*y + 3*y = x^3 + 2*x^2 + 4*x + 5
                    over Finite Field of size 65537
              To:   Elliptic Curve defined by y^2 + x*y + 3*y = x^3 + 2*x^2 + 28339*x + 59518
                    over Finite Field of size 65537
            sage: psi = phi.dual(); psi
            Composite morphism of degree 9 = 3^2:
              From: Elliptic Curve defined by y^2 + x*y + 3*y = x^3 + 2*x^2 + 28339*x + 59518
                    over Finite Field of size 65537
              To:   Elliptic Curve defined by y^2 + x*y + 3*y = x^3 + 2*x^2 + 4*x + 5
                    over Finite Field of size 65537
            sage: psi * phi == phi.domain().scalar_multiplication(phi.degree())         # needs sage.symbolic
            True
            sage: phi * psi == psi.domain().scalar_multiplication(psi.degree())         # needs sage.symbolic
            True
        """
        phis = (phi.dual() for phi in self._phis[::-1])
        return EllipticCurveHom_composite.from_factors(phis)

    def formal(self, prec=20):
        """
        Return the formal isogeny corresponding to this composite
        isogeny as a power series in the variable `t=-x/y` on the
        domain curve.

        EXAMPLES::

            sage: # needs sage.rings.finite_rings
            sage: from sage.schemes.elliptic_curves.hom_composite import EllipticCurveHom_composite
            sage: E = EllipticCurve(GF(65537), [1,2,3,4,5])
            sage: P = E.lift_x(7321)
            sage: phi = EllipticCurveHom_composite(E, P)
            sage: phi.formal()
            t + 54203*t^5 + 48536*t^6 + 40698*t^7 + 37808*t^8 + 21111*t^9 + 42381*t^10
             + 46688*t^11 + 657*t^12 + 38916*t^13 + 62261*t^14 + 59707*t^15
             + 30767*t^16 + 7248*t^17 + 60287*t^18 + 50451*t^19 + 38305*t^20
             + 12312*t^21 + 31329*t^22 + O(t^23)
            sage: (phi.dual() * phi).formal(prec=5)
            9*t + 65501*t^2 + 65141*t^3 + 59183*t^4 + 21491*t^5 + 8957*t^6
             + 999*t^7 + O(t^8)
        """
        res = self._phis[-1].formal(prec=prec)
        for phi in self._phis[:-1][::-1]:
            res = res(phi.formal(prec=prec))
        return res

    def scaling_factor(self):
        r"""
        Return the Weierstrass scaling factor associated to this
        composite morphism.

        The scaling factor is the constant `u` (in the base field)
        such that `\varphi^* \omega_2 = u \omega_1`, where
        `\varphi: E_1\to E_2` is this morphism and `\omega_i` are
        the standard Weierstrass differentials on `E_i` defined by
        `\mathrm dx/(2y+a_1x+a_3)`.

        EXAMPLES::

            sage: # needs sage.rings.finite_rings
            sage: from sage.schemes.elliptic_curves.hom_composite import EllipticCurveHom_composite
            sage: from sage.schemes.elliptic_curves.weierstrass_morphism import WeierstrassIsomorphism
            sage: E = EllipticCurve(GF(65537), [1,2,3,4,5])
            sage: P = E.lift_x(7321)
            sage: phi = EllipticCurveHom_composite(E, P)
            sage: phi = WeierstrassIsomorphism(phi.codomain(), [7,8,9,10]) * phi
            sage: phi.formal()
            7*t + 65474*t^2 + 511*t^3 + 61316*t^4 + 20548*t^5 + 45511*t^6 + 37285*t^7
             + 48414*t^8 + 9022*t^9 + 24025*t^10 + 35986*t^11 + 55397*t^12 + 25199*t^13
             + 18744*t^14 + 46142*t^15 + 9078*t^16 + 18030*t^17 + 47599*t^18
             + 12158*t^19 + 50630*t^20 + 56449*t^21 + 43320*t^22 + O(t^23)
            sage: phi.scaling_factor()
            7

        ALGORITHM: The scaling factor is multiplicative under
        composition, so we return the product of the individual
        scaling factors associated to each factor.
        """
        return prod(phi.scaling_factor() for phi in self._phis)

    def inseparable_degree(self):
        r"""
        Return the inseparable degree of this morphism.

        Like the degree, the inseparable degree is multiplicative
        under composition, so this method returns the product of
        the inseparable degrees of the factors.

        EXAMPLES::

            sage: E = EllipticCurve(j=GF(11^5).random_element())
            sage: phi = E.frobenius_isogeny(2) * E.scalar_multiplication(77)
            sage: type(phi)
            <class 'sage.schemes.elliptic_curves.hom_composite.EllipticCurveHom_composite'>
            sage: phi.inseparable_degree()
            1331
        """
        return prod(phi.inseparable_degree() for phi in self._phis)

    @property
    def _rest(self):
        """
        Internal property such that ``self == self._rest * self._phis[0]``.

        TESTS::

            sage: E.<P, Q> = EllipticCurve(GF(5^2), [1, 2, 3, 3, 1])
            sage: f = E.isogeny([P*3, Q*3])
            sage: assert f == f._rest * f._phis[0]
        """
        return EllipticCurveHom_composite.from_factors(self._phis[1:], strict=False)

    def kernel_points(self):
        """
        Return an iterator over the points in the kernel of this
        elliptic-curve morphism.

        EXAMPLES::

            sage: E.<P, Q> = EllipticCurve(GF(5^2), [1, 2, 3, 3, 1])
            sage: f = E.isogeny([P*3, Q*3])
            sage: f
            Composite morphism of degree 4 = 2^2:
              From: Elliptic Curve defined by y^2 + x*y + 3*y = x^3 + 2*x^2 + 3*x + 1 over Finite Field in z2 of size 5^2
              To:   Elliptic Curve defined by y^2 + x*y + 3*y = x^3 + 2*x^2 + 3*x + 3 over Finite Field in z2 of size 5^2
            sage: set(f.kernel_points())
            {(0 : 1 : 0), (4 : 4 : 1), (2*z2 + 4 : 4*z2 + 4 : 1), (3*z2 + 1 : z2 + 3 : 1)}
        """
        yield from self.inverse_image(self.codomain().zero(), all=True)

    def inverse_image(self, Q, /, *, all=False):
        """
        Return an arbitrary element ``P`` in the domain such that
        ``self(P) == Q``, or raise ``ValueError`` if no such
        element exists.

        INPUT:

        - ``Q`` -- a point
        - ``all`` -- boolean; if ``True``, returns an iterator over all points
          in the inverse image

        EXAMPLES::

            sage: E.<P, Q> = EllipticCurve(GF(5^2), [1, 2, 3, 3, 1])
            sage: f = E.isogeny([P*3, Q*3])
            sage: f
            Composite morphism of degree 4 = 2^2:
              From: Elliptic Curve defined by y^2 + x*y + 3*y = x^3 + 2*x^2 + 3*x + 1 over Finite Field in z2 of size 5^2
              To:   Elliptic Curve defined by y^2 + x*y + 3*y = x^3 + 2*x^2 + 3*x + 3 over Finite Field in z2 of size 5^2
            sage: f(f.inverse_image(f(Q))) == f(Q)
            True
            sage: E.scalar_multiplication(-1).inverse_image(P) == -P
            True
            sage: f.inverse_image(f.codomain().0)
            Traceback (most recent call last):
            ...
            ValueError...
            sage: len(list(f.inverse_image(f(Q), all=True)))
            4

        Test a large example. It should finish in a few seconds::

            sage: p = 3 * 2^143 - 1
            sage: GF(p^2).inject_variables()
            Defining z2
            sage: E = EllipticCurve(GF(p^2), [1,0])
            sage: P = E.lift_x(31415926535897932384626433832795028841971 - z2)
            sage: f = E.isogeny(P, algorithm="factored")
            sage: Q = f(E.lift_x(2718281828459045235360287471352662497757 - z2)); Q
            (14253459515090351074737629944491750308703143*z2 + 17548601963968266930680314841240982076784493 : ... : 1)
            sage: f.inverse_image(Q)  # long time
            (...)

        TESTS:

        Normally, a :class:`EllipticCurveHom_composite` has ``len(self._phis) > 1``,
        but if :meth:`from_factors` is called with ``strict=True``, or if the user
        constructs a :class:`EllipticCurveHom_composite` object directly, then it is
        possible to violate this condition. We test for this case::

            sage: E.<P> = EllipticCurve(GF(5), [1, 2])
            sage: from sage.schemes.elliptic_curves.hom_composite import EllipticCurveHom_composite
            sage: f = EllipticCurveHom_composite(E, P*2); f
            Composite morphism of degree 2:
              From: Elliptic Curve defined by y^2 = x^3 + x + 2 over Finite Field of size 5
              To:   Elliptic Curve defined by y^2 = x^3 + x over Finite Field of size 5
            sage: len(f._phis)
            1
            sage: f(f.inverse_image(f(P))) == f(P)
            True
            sage: set(f.inverse_image(f(P), all=True))
            {(1 : 2 : 1), (1 : 3 : 1)}

        The current implementation guarantees :attr:`_phis` is not empty::

            sage: f = EllipticCurveHom_composite.from_factors((), E); f
            Composite morphism of degree 1:
              From: Elliptic Curve defined by y^2 = x^3 + x + 2 over Finite Field of size 5
              To:   Elliptic Curve defined by y^2 = x^3 + x + 2 over Finite Field of size 5
            sage: len(f._phis)
            1
            sage: f.inverse_image(P) == P
            True
        """
        if len(self._phis) == 1:
            return self._phis[0].inverse_image(Q, all=all)
        if all:
            return (R for P in self._rest.inverse_image(Q, all=True) for R in self._phis[0].inverse_image(P, all=True))
        try:
            return next(self.inverse_image(Q, all=True))
        except StopIteration:
            raise ValueError

    def push_subgroup(self, f):
        r"""
        Given a minimal polynomial (see :meth:`~EllipticCurveHom.minimal_polynomial`)
        of a subgroup `G` of the domain curve of this isogeny, return a minimal polynomial
        of the image of `G` under this isogeny.

        ALGORITHM: iterative :meth:`EllipticCurveHom.push_subgroup()`

        EXAMPLES::

            sage: E = EllipticCurve(GF((2^61-1, 2)), [1,0])
            sage: phi = next(E.isogenies_degree(7)); phi
            Isogeny of degree 7
              from Elliptic Curve defined by y^2 = x^3 + x over Finite Field in z2 of size 2305843009213693951^2
              to Elliptic Curve defined by y^2 = x^3 + (595688734420561721*z2+584021682365204922)*x + (2058397526093132314*z2+490140893682260802) over Finite Field in z2 of size 2305843009213693951^2
            sage: psi = E.isogeny(E.lift_x(48), algorithm='factored'); psi
            Composite morphism of degree 36028797018963968 = 2^55:
              From: Elliptic Curve defined by y^2 = x^3 + x over Finite Field in z2 of size 2305843009213693951^2
              To:   Elliptic Curve defined by y^2 = x^3 + 938942632807894005*x + 1238942515234646252 over Finite Field in z2 of size 2305843009213693951^2
            sage: f = phi.minimal_polynomial()
            sage: g = psi.push_subgroup(f)
            sage: h = psi.codomain().kernel_polynomial_from_divisor(g, phi.degree())
            sage: chi = psi.codomain().isogeny(h); chi
            Isogeny of degree 7 from Elliptic Curve defined by y^2 = x^3 + 938942632807894005*x + 1238942515234646252 over Finite Field in z2 of size 2305843009213693951^2 to Elliptic Curve defined by y^2 = x^3 + (1406897314822267524*z2+1659665944678449850)*x + (650305521764753329*z2+1047269804324934563) over Finite Field in z2 of size 2305843009213693951^2
            sage: x = phi.kernel_polynomial().any_root()
            sage: K = E.change_ring(E.base_field().extension(2)).lift_x(x)
            sage: (chi * psi)._eval(K)
            (0 : 1 : 0)
        """
        for phi in self.factors():
            f = phi.push_subgroup(f)
        return f



# sage_setup: distribution = sagemath-modules
r"""
Exterior powers of free modules

Given a free module `M` of finite rank over a commutative ring `R`
and a positive integer `p`, the `p`-*th exterior power of* `M`
is the set `\Lambda^p(M)` of all alternating contravariant tensors of
degree `p` on `M`, i.e. of all multilinear maps

.. MATH::

    \underbrace{M^*\times\cdots\times M^*}_{p\ \; \mbox{times}}
    \longrightarrow R

that vanish whenever any of two of their arguments are equal
(`M^*` stands for the dual of `M`).
Note that `\Lambda^1(M) = M`. The exterior power
`\Lambda^p(M)` is a free module of rank `\binom{n}{p}` over `R`,
where `n` is the rank of `M`.

Similarly, the `p`-*th exterior power of the dual of* `M`
is the set `\Lambda^p(M^*)` of all alternating forms of degree `p` on
`M`, i.e. of all multilinear maps

.. MATH::

    \underbrace{M\times\cdots\times M}_{p\ \; \mbox{times}}
    \longrightarrow R

that vanish whenever any of two of their arguments are equal.
Note that `\Lambda^1(M^*) = M^*` (the dual of `M`). The exterior power
`\Lambda^p(M^*)` is a free module of rank `\binom{n}{p}` over `R`,
where `n` is the rank of `M`.

The class :class:`ExtPowerFreeModule` implements `\Lambda^p(M)`, while
the class :class:`ExtPowerDualFreeModule` implements `\Lambda^p(M^*)`.

AUTHORS:

- Eric Gourgoulhon: initial version, regarding `\Lambda^p(M^*)` only
  (2015); add class for `\Lambda^p(M)` (2017)


REFERENCES:

- \K. Conrad: *Exterior powers* [Con2013]_
- Chap. 19 of S. Lang: *Algebra* [Lan2002]_
"""
# ****************************************************************************
#       Copyright (C) 2017 Eric Gourgoulhon <eric.gourgoulhon@obspm.fr>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#  as published by the Free Software Foundation; either version 2 of
#  the License, or (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.misc.cachefunc import cached_method
from sage.rings.integer import Integer
from sage.rings.integer_ring import ZZ
from sage.tensor.modules.finite_rank_free_module import FiniteRankFreeModule_abstract
from sage.tensor.modules.free_module_tensor import FreeModuleTensor
from sage.tensor.modules.alternating_contr_tensor import AlternatingContrTensor
from sage.tensor.modules.free_module_alt_form import FreeModuleAltForm


class ExtPowerFreeModule(FiniteRankFreeModule_abstract):
    r"""
    Exterior power of a free module of finite rank over a commutative
    ring.

    Given a free module `M` of finite rank over a commutative ring `R`
    and a positive integer `p`, the `p`-*th exterior power of* `M` is
    the set `\Lambda^p(M)` of all alternating contravariant tensors of
    degree `p` on `M`, i.e. of all multilinear maps

    .. MATH::

        \underbrace{M^*\times\cdots\times M^*}_{p\ \; \mbox{times}}
        \longrightarrow R

    that vanish whenever any of two of their arguments are equal.
    Note that `\Lambda^1(M) = M`.

    `\Lambda^p(M)` is a free module of rank `\binom{n}{p}` over
    `R`, where `n` is the rank of `M`.
    Accordingly, the class :class:`ExtPowerFreeModule` inherits from the
    class
    :class:`~sage.tensor.modules.finite_rank_free_module.FiniteRankFreeModule_abstract`.

    This is a Sage *parent* class, whose *element* class is
    :class:`~sage.tensor.modules.alternating_contr_tensor.AlternatingContrTensor`

    INPUT:

    - ``fmodule`` -- free module `M` of finite rank, as an instance of
      :class:`~sage.tensor.modules.finite_rank_free_module.FiniteRankFreeModule`
    - ``degree`` -- positive integer; the degree `p` of the alternating
      elements
    - ``name`` -- (default: ``None``) string; name given to `\Lambda^p(M)`
    - ``latex_name`` -- (default: ``None``) string; LaTeX symbol to
      denote `\Lambda^p(M)`

    EXAMPLES:

    2nd exterior power of the dual of a free `\ZZ`-module of rank 3::

        sage: M = FiniteRankFreeModule(ZZ, 3, name='M')
        sage: e = M.basis('e')
        sage: from sage.tensor.modules.ext_pow_free_module import ExtPowerFreeModule
        sage: A = ExtPowerFreeModule(M, 2) ; A
        2nd exterior power of the Rank-3 free module M over the
         Integer Ring

    Instead of importing ExtPowerFreeModule in the global name space, it is
    recommended to use the module's method
    :meth:`~sage.tensor.modules.finite_rank_free_module.FiniteRankFreeModule.exterior_power`::

        sage: A = M.exterior_power(2) ; A
        2nd exterior power of the Rank-3 free module M over the
         Integer Ring
        sage: latex(A)
        \Lambda^{2}\left(M\right)

    ``A`` is a module (actually a free module) over `\ZZ`::

        sage: A.category()
        Category of finite dimensional modules over Integer Ring
        sage: A in Modules(ZZ)
        True
        sage: A.rank()
        3
        sage: A.base_ring()
        Integer Ring
        sage: A.base_module()
        Rank-3 free module M over the Integer Ring

    ``A`` is a *parent* object, whose elements are alternating
    contravariant tensors, represented by instances of the class
    :class:`~sage.tensor.modules.alternating_contr_tensor.AlternatingContrTensor`::

        sage: a = A.an_element() ; a
        Alternating contravariant tensor of degree 2 on the Rank-3 free
         module M over the Integer Ring
        sage: a.display() # expansion with respect to M's default basis (e)
        e_0∧e_1
        sage: from sage.tensor.modules.alternating_contr_tensor import AlternatingContrTensor
        sage: isinstance(a, AlternatingContrTensor)
        True
        sage: a in A
        True
        sage: A.is_parent_of(a)
        True

    Elements can be constructed from ``A``. In particular, 0 yields
    the zero element of ``A``::

        sage: A(0)
        Alternating contravariant tensor zero of degree 2 on the Rank-3
         free module M over the Integer Ring
        sage: A(0) is A.zero()
        True

    while nonzero elements are constructed by providing their components in a
    given basis::

        sage: e
        Basis (e_0,e_1,e_2) on the Rank-3 free module M over the Integer Ring
        sage: comp = [[0,3,-1],[-3,0,4],[1,-4,0]]
        sage: a = A(comp, basis=e, name='a') ; a
        Alternating contravariant tensor a of degree 2 on the Rank-3
         free module M over the Integer Ring
        sage: a.display(e)
        a = 3 e_0∧e_1 - e_0∧e_2 + 4 e_1∧e_2

    An alternative is to construct the alternating contravariant tensor from an
    empty list of components and to set the nonzero components afterwards::

        sage: a = A([], name='a')
        sage: a.set_comp(e)[0,1] = 3
        sage: a.set_comp(e)[0,2] = -1
        sage: a.set_comp(e)[1,2] = 4
        sage: a.display(e)
        a = 3 e_0∧e_1 - e_0∧e_2 + 4 e_1∧e_2

    The exterior powers are unique::

        sage: A is M.exterior_power(2)
        True

    The exterior power `\Lambda^1(M)` is nothing but `M`::

        sage: M.exterior_power(1) is M
        True

    For a degree `p\geq 2`, there is a coercion
    `\Lambda^p(M)\rightarrow T^{(p,0)}(M)`::

        sage: T20 = M.tensor_module(2,0) ; T20
        Free module of type-(2,0) tensors on the Rank-3 free module M
         over the Integer Ring
        sage: T20.has_coerce_map_from(A)
        True

    Of course, there is no coercion in the reverse direction::

        sage: A.has_coerce_map_from(T20)
        False

    The coercion map `\Lambda^2(M)\rightarrow T^{(2,0)}(M)` in action::

        sage: ta = T20(a) ; ta
        Type-(2,0) tensor a on the Rank-3 free module M over the Integer Ring
        sage: ta.display(e)
        a = 3 e_0⊗e_1 - e_0⊗e_2 - 3 e_1⊗e_0 + 4 e_1⊗e_2 + e_2⊗e_0 - 4 e_2⊗e_1
        sage: a.display(e)
        a = 3 e_0∧e_1 - e_0∧e_2 + 4 e_1∧e_2
        sage: ta.symmetries()  # the antisymmetry is of course preserved
        no symmetry;  antisymmetry: (0, 1)
        sage: ta == a  # equality as type-(2,0) tensors
        True
    """

    Element = AlternatingContrTensor

    def __init__(self, fmodule, degree, name=None, latex_name=None):
        r"""
        TESTS::

            sage: from sage.tensor.modules.ext_pow_free_module import ExtPowerFreeModule
            sage: M = FiniteRankFreeModule(ZZ, 3, name='M')
            sage: e = M.basis('e')
            sage: A = ExtPowerFreeModule(M, 2) ; A
            2nd exterior power of the Rank-3 free module M over the
             Integer Ring
            sage: TestSuite(A).run()
        """
        from sage.arith.misc import binomial
        from sage.typeset.unicode_characters import unicode_bigwedge
        self._fmodule = fmodule
        self._degree = ZZ(degree)
        rank = binomial(fmodule._rank, degree)
        if name is None and fmodule._name is not None:
            name = unicode_bigwedge + r'^{}('.format(degree) \
                   + fmodule._name + ')'
        if latex_name is None and fmodule._latex_name is not None:
            latex_name = r'\Lambda^{' + str(degree) + r'}\left(' \
                         + fmodule._latex_name + r'\right)'
        super().__init__(fmodule._ring, rank,
                         name=name, latex_name=latex_name)
        fmodule._all_modules.add(self)

    def construction(self):
        r"""
        Return the functorial construction of ``self``.

        This implementation just returns ``None``, as no functorial construction is implemented.

        TESTS::

            sage: from sage.tensor.modules.ext_pow_free_module import ExtPowerFreeModule
            sage: M = FiniteRankFreeModule(ZZ, 3, name='M')
            sage: e = M.basis('e')
            sage: A = ExtPowerFreeModule(M, 2)
            sage: A.construction() is None
            True
        """
        # No construction until https://github.com/sagemath/sage/issues/30242
        # makes this a quotient of TensorFreeModule
        return None

    #### Parent methods

    def _element_constructor_(self, comp=[], basis=None, name=None,
                              latex_name=None):
        r"""
        Construct an alternating contravariant tensor.

        EXAMPLES::

            sage: M = FiniteRankFreeModule(ZZ, 3, name='M')
            sage: e = M.basis('e')
            sage: A = M.exterior_power(2)
            sage: a = A._element_constructor_(0) ; a
            Alternating contravariant tensor zero of degree 2 on the
             Rank-3 free module M over the Integer Ring
            sage: a = A._element_constructor_([], name='a') ; a
            Alternating contravariant tensor a of degree 2 on the Rank-3
             free module M over the Integer Ring
            sage: a[e,0,2], a[e,1,2] = 3, -1
            sage: a.display()
            a = 3 e_0∧e_2 - e_1∧e_2
        """
        if isinstance(comp, (int, Integer)) and comp == 0:
            return self.zero()
        resu = self.element_class(self._fmodule, self._degree, name=name,
                                  latex_name=latex_name)
        if comp:
            resu.set_comp(basis)[:] = comp
        return resu

    def _an_element_(self):
        r"""
        Construct some (unnamed) alternating contravariant tensor.

        EXAMPLES::

            sage: M = FiniteRankFreeModule(QQ, 4, name='M')
            sage: e = M.basis('e')
            sage: a = M.exterior_power(2)._an_element_() ; a
            Alternating contravariant tensor of degree 2 on the 4-dimensional vector space M
             over the Rational Field
            sage: a.display()
            1/2 e_0∧e_1
            sage: a = M.exterior_power(3)._an_element_() ; a
            Alternating contravariant tensor of degree 3 on the 4-dimensional vector space M
             over the Rational Field
            sage: a.display()
            1/2 e_0∧e_1∧e_2
            sage: a = M.exterior_power(4)._an_element_() ; a
            Alternating contravariant tensor of degree 4 on the 4-dimensional vector space M
             over the Rational Field
            sage: a.display()
            1/2 e_0∧e_1∧e_2∧e_3

        TESTS:

        When the base module has no default basis, a default
        basis will be set for it::

            sage: M2 = FiniteRankFreeModule(QQ, 4, name='M2')
            sage: a = M2.exterior_power(2)._an_element_(); a
            Alternating contravariant tensor of degree 2
            on the 4-dimensional vector space M2 over the Rational Field
            sage: a + a
            Alternating contravariant tensor of degree 2
            on the 4-dimensional vector space M2 over the Rational Field
            sage: M2.default_basis()
            Basis (e_0,e_1,e_2,e_3) on the 4-dimensional vector space M2 over the Rational Field
        """
        resu = self.element_class(self._fmodule, self._degree)
        # Make sure that the base module has a default basis
        self._fmodule.an_element()
        sindex = self._fmodule._sindex
        ind = [sindex + i for i in range(resu._tensor_rank)]
        resu.set_comp()[ind] = self._fmodule._ring.an_element()
        return resu

    #### End of parent methods

    @cached_method
    def zero(self):
        r"""
        Return the zero of ``self``.

        EXAMPLES::

            sage: M = FiniteRankFreeModule(ZZ, 3, name='M')
            sage: e = M.basis('e')
            sage: A = M.exterior_power(2)
            sage: A.zero()
            Alternating contravariant tensor zero of degree 2 on the Rank-3 free
             module M over the Integer Ring
            sage: A(0) is A.zero()
            True
        """
        resu = self._element_constructor_(name='zero', latex_name='0')
        for basis in self._fmodule._known_bases:
            resu._add_comp_unsafe(basis)
            # (since new components are initialized to zero)
        resu._is_zero = True # This element is certainly zero
        resu.set_immutable()
        return resu

    def _repr_(self):
        r"""
        Return a string representation of ``self``.

        EXAMPLES::

            sage: M = FiniteRankFreeModule(ZZ, 5, name='M')
            sage: M.exterior_power(2)._repr_()
            '2nd exterior power of the Rank-5 free module M over the Integer Ring'
            sage: M.exterior_power(3)._repr_()
            '3rd exterior power of the Rank-5 free module M over the Integer Ring'
            sage: M.exterior_power(4)._repr_()
            '4th exterior power of the Rank-5 free module M over the Integer Ring'
            sage: M.exterior_power(5)._repr_()
            '5th exterior power of the Rank-5 free module M over the Integer Ring'
            sage: M.exterior_power(21)._repr_()
            '21st exterior power of the Rank-5 free module M over the Integer Ring'
        """
        description = "{}".format(self._degree.ordinal_str())
        description += " exterior power of the {}".format(self._fmodule)
        return description

    def base_module(self):
        r"""
        Return the free module on which ``self`` is constructed.

        OUTPUT:

        - instance of :class:`FiniteRankFreeModule` representing the
          free module on which the exterior power is defined.

        EXAMPLES::

            sage: M = FiniteRankFreeModule(ZZ, 5, name='M')
            sage: A = M.exterior_power(2)
            sage: A.base_module()
            Rank-5 free module M over the Integer Ring
            sage: A.base_module() is M
            True
        """
        return self._fmodule

    def degree(self):
        r"""
        Return the degree of ``self``.

        OUTPUT:

        - integer `p` such that ``self`` is the exterior power
          `\Lambda^p(M)`

        EXAMPLES::

            sage: M = FiniteRankFreeModule(ZZ, 5, name='M')
            sage: A = M.exterior_power(2)
            sage: A.degree()
            2
            sage: M.exterior_power(4).degree()
            4
        """
        return self._degree

#***********************************************************************


class ExtPowerDualFreeModule(FiniteRankFreeModule_abstract):
    r"""
    Exterior power of the dual of a free module of finite rank
    over a commutative ring.

    Given a free module `M` of finite rank over a commutative ring `R`
    and a positive integer `p`, the `p`-*th exterior power of the dual of*
    `M` is the set `\Lambda^p(M^*)` of all alternating forms of degree
    `p` on `M`, i.e. of all multilinear maps

    .. MATH::

        \underbrace{M\times\cdots\times M}_{p\ \; \mbox{times}}
        \longrightarrow R

    that vanish whenever any of two of their arguments are equal.
    Note that `\Lambda^1(M^*) = M^*` (the dual of `M`).

    `\Lambda^p(M^*)` is a free module of rank `\binom{n}{p}` over
    `R`, where `n` is the rank of `M`.
    Accordingly, the class :class:`ExtPowerDualFreeModule` inherits from
    the class
    :class:`~sage.tensor.modules.finite_rank_free_module.FiniteRankFreeModule_abstract`.

    This is a Sage *parent* class, whose *element* class is
    :class:`~sage.tensor.modules.free_module_alt_form.FreeModuleAltForm`.

    INPUT:

    - ``fmodule`` -- free module `M` of finite rank, as an instance of
      :class:`~sage.tensor.modules.finite_rank_free_module.FiniteRankFreeModule`
    - ``degree`` -- positive integer; the degree `p` of the alternating
      forms
    - ``name`` -- (default: ``None``) string; name given to `\Lambda^p(M^*)`
    - ``latex_name`` -- (default: ``None``) string; LaTeX symbol to
      denote `\Lambda^p(M^*)`

    EXAMPLES:

    2nd exterior power of the dual of a free `\ZZ`-module of rank 3::

        sage: M = FiniteRankFreeModule(ZZ, 3, name='M')
        sage: e = M.basis('e')
        sage: from sage.tensor.modules.ext_pow_free_module import ExtPowerDualFreeModule
        sage: A = ExtPowerDualFreeModule(M, 2) ; A
        2nd exterior power of the dual of the Rank-3 free module M over the
         Integer Ring

    Instead of importing ExtPowerDualFreeModule in the global name space,
    it is recommended to use the module's method
    :meth:`~sage.tensor.modules.finite_rank_free_module.FiniteRankFreeModule.dual_exterior_power`::

        sage: A = M.dual_exterior_power(2) ; A
        2nd exterior power of the dual of the Rank-3 free module M over the
         Integer Ring
        sage: latex(A)
        \Lambda^{2}\left(M^*\right)

    ``A`` is a module (actually a free module) over `\ZZ`::

        sage: A.category()
        Category of finite dimensional modules over Integer Ring
        sage: A in Modules(ZZ)
        True
        sage: A.rank()
        3
        sage: A.base_ring()
        Integer Ring
        sage: A.base_module()
        Rank-3 free module M over the Integer Ring

    ``A`` is a *parent* object, whose elements are alternating forms,
    represented by instances of the class
    :class:`~sage.tensor.modules.free_module_alt_form.FreeModuleAltForm`::

        sage: a = A.an_element() ; a
        Alternating form of degree 2 on the Rank-3 free module M over the
         Integer Ring
        sage: a.display() # expansion with respect to M's default basis (e)
        e^0∧e^1
        sage: from sage.tensor.modules.free_module_alt_form import FreeModuleAltForm
        sage: isinstance(a, FreeModuleAltForm)
        True
        sage: a in A
        True
        sage: A.is_parent_of(a)
        True

    Elements can be constructed from ``A``. In particular, 0 yields
    the zero element of ``A``::

        sage: A(0)
        Alternating form zero of degree 2 on the Rank-3 free module M over the
         Integer Ring
        sage: A(0) is A.zero()
        True

    while nonzero elements are constructed by providing their components in a
    given basis::

        sage: e
        Basis (e_0,e_1,e_2) on the Rank-3 free module M over the Integer Ring
        sage: comp = [[0,3,-1],[-3,0,4],[1,-4,0]]
        sage: a = A(comp, basis=e, name='a') ; a
        Alternating form a of degree 2 on the Rank-3 free module M over the
         Integer Ring
        sage: a.display(e)
        a = 3 e^0∧e^1 - e^0∧e^2 + 4 e^1∧e^2

    An alternative is to construct the alternating form from an empty list of
    components and to set the nonzero components afterwards::

        sage: a = A([], name='a')
        sage: a.set_comp(e)[0,1] = 3
        sage: a.set_comp(e)[0,2] = -1
        sage: a.set_comp(e)[1,2] = 4
        sage: a.display(e)
        a = 3 e^0∧e^1 - e^0∧e^2 + 4 e^1∧e^2

    The exterior powers are unique::

        sage: A is M.dual_exterior_power(2)
        True

    The exterior power `\Lambda^1(M^*)` is nothing but `M^*`::

        sage: M.dual_exterior_power(1) is M.dual()
        True
        sage: M.dual()
        Dual of the Rank-3 free module M over the Integer Ring
        sage: latex(M.dual())
        M^*

    It also coincides with the module of type-`(0,1)` tensors::

        sage: M.dual_exterior_power(1) is M.tensor_module(0,1)
        True

    For a degree `p\geq 2`, there is a coercion map
    `\Lambda^p(M^*)\rightarrow T^{(0,p)}(M)`::

        sage: T02 = M.tensor_module(0,2) ; T02
        Free module of type-(0,2) tensors on the Rank-3 free module M over the
         Integer Ring
        sage: T02.has_coerce_map_from(A)
        True
        sage: A.has_coerce_map_from(T02)
        False

    The coercion map `\Lambda^2(M^*)\rightarrow T^{(0,2)}(M)` in action::

        sage: ta = T02(a) ; ta
        Type-(0,2) tensor a on the Rank-3 free module M over the Integer Ring
        sage: ta.display(e)
        a = 3 e^0⊗e^1 - e^0⊗e^2 - 3 e^1⊗e^0 + 4 e^1⊗e^2 + e^2⊗e^0 - 4 e^2⊗e^1
        sage: a.display(e)
        a = 3 e^0∧e^1 - e^0∧e^2 + 4 e^1∧e^2
        sage: ta.symmetries() # the antisymmetry is of course preserved
        no symmetry;  antisymmetry: (0, 1)
    """

    Element = FreeModuleAltForm

    def __init__(self, fmodule, degree, name=None, latex_name=None):
        r"""
        TESTS::

            sage: from sage.tensor.modules.ext_pow_free_module import ExtPowerDualFreeModule
            sage: M = FiniteRankFreeModule(ZZ, 3, name='M')
            sage: e = M.basis('e')
            sage: A = ExtPowerDualFreeModule(M, 2) ; A
            2nd exterior power of the dual of the Rank-3 free module M over
             the Integer Ring
            sage: TestSuite(A).run()
        """
        from sage.arith.misc import binomial
        from sage.typeset.unicode_characters import unicode_bigwedge
        self._fmodule = fmodule
        self._degree = ZZ(degree)
        rank = binomial(fmodule._rank, degree)
        if name is None and fmodule._name is not None:
            name = unicode_bigwedge + r'^{}('.format(degree) \
                   + fmodule._name + '*)'
        if latex_name is None and fmodule._latex_name is not None:
            latex_name = r'\Lambda^{' + str(degree) + r'}\left(' \
                         + fmodule._latex_name + r'^*\right)'
        super().__init__(fmodule._ring, rank, name=name,
                         latex_name=latex_name)
        fmodule._all_modules.add(self)

    def construction(self):
        r"""
        Return the functorial construction of ``self``.

        This implementation just returns ``None``, as no functorial construction is implemented.

        TESTS::

            sage: from sage.tensor.modules.ext_pow_free_module import ExtPowerDualFreeModule
            sage: M = FiniteRankFreeModule(ZZ, 3, name='M')
            sage: e = M.basis('e')
            sage: A = ExtPowerDualFreeModule(M, 2)
            sage: A.construction() is None
            True
        """
        # No construction until https://github.com/sagemath/sage/issues/30242
        # makes this a quotient of TensorFreeModule
        return None

    #### Parent methods

    def _element_constructor_(self, comp=[], basis=None, name=None,
                              latex_name=None):
        r"""
        Construct an alternating form.

        EXAMPLES::

            sage: M = FiniteRankFreeModule(ZZ, 3, name='M')
            sage: e = M.basis('e')
            sage: A = M.dual_exterior_power(2)
            sage: a = A._element_constructor_(0) ; a
            Alternating form zero of degree 2 on the Rank-3 free module M over
             the Integer Ring
            sage: a = A._element_constructor_([], name='a') ; a
            Alternating form a of degree 2 on the Rank-3 free module M over
             the Integer Ring
            sage: a[e,0,2], a[e,1,2] = 3, -1
            sage: a.display()
            a = 3 e^0∧e^2 - e^1∧e^2
        """
        if isinstance(comp, (int, Integer)) and comp == 0:
            return self.zero()
        if isinstance(comp, FreeModuleTensor):
            # coercion of a tensor of type (0,1) to a linear form
            tensor = comp # for readability
            if tensor.tensor_type() == (0,1) and self._degree == 1 and \
                                         tensor.base_module() is self._fmodule:
                resu = self.element_class(self._fmodule, 1, name=tensor._name,
                                          latex_name=tensor._latex_name)
                for basis, comp in tensor._components.items():
                    resu._components[basis] = comp.copy()
                return resu
            else:
                raise TypeError("cannot coerce the {} ".format(tensor) +
                                "to an element of {}".format(self))
        # standard construction
        resu = self.element_class(self._fmodule, self._degree, name=name,
                                  latex_name=latex_name)
        if comp:
            resu.set_comp(basis)[:] = comp
        return resu

    def _an_element_(self):
        r"""
        Construct some (unnamed) alternating form.

        EXAMPLES::

            sage: M = FiniteRankFreeModule(QQ, 4, name='M')
            sage: e = M.basis('e')
            sage: a = M.dual_exterior_power(2)._an_element_() ; a
            Alternating form of degree 2 on the 4-dimensional vector space M
             over the Rational Field
            sage: a.display()
            1/2 e^0∧e^1
            sage: a = M.dual_exterior_power(3)._an_element_() ; a
            Alternating form of degree 3 on the 4-dimensional vector space M
             over the Rational Field
            sage: a.display()
            1/2 e^0∧e^1∧e^2
            sage: a = M.dual_exterior_power(4)._an_element_() ; a
            Alternating form of degree 4 on the 4-dimensional vector space M
             over the Rational Field
            sage: a.display()
            1/2 e^0∧e^1∧e^2∧e^3

        TESTS:

        When the base module has no default basis, a default
        basis will be set for it::

            sage: M2 = FiniteRankFreeModule(QQ, 4, name='M2')
            sage: a = M2.dual_exterior_power(2)._an_element_(); a
            Alternating form of degree 2 on the 4-dimensional vector space M2 over the Rational Field
            sage: a + a
            Alternating form of degree 2 on the 4-dimensional vector space M2 over the Rational Field
            sage: M2.default_basis()
            Basis (e_0,e_1,e_2,e_3) on the 4-dimensional vector space M2 over the Rational Field
        """
        resu = self.element_class(self._fmodule, self._degree)
        # Make sure that the base module has a default basis
        self._fmodule.an_element()
        sindex = self._fmodule._sindex
        ind = [sindex + i for i in range(resu._tensor_rank)]
        resu.set_comp()[ind] = self._fmodule._ring.an_element()
        return resu

    #### End of parent methods

    @cached_method
    def zero(self):
        r"""
        Return the zero of ``self``.

        EXAMPLES::

            sage: M = FiniteRankFreeModule(ZZ, 3, name='M')
            sage: e = M.basis('e')
            sage: A = M.dual_exterior_power(2)
            sage: A.zero()
            Alternating form zero of degree 2 on the Rank-3 free module M over
             the Integer Ring
            sage: A(0) is A.zero()
            True
        """
        resu = self._element_constructor_(name='zero', latex_name='0')
        for basis in self._fmodule._known_bases:
            resu._components[basis] = resu._new_comp(basis)
            # (since new components are initialized to zero)
        resu._is_zero = True # This element is certainly zero
        resu.set_immutable()
        return resu

    def _repr_(self):
        r"""
        Return a string representation of ``self``.

        EXAMPLES::

            sage: M = FiniteRankFreeModule(ZZ, 5, name='M')
            sage: M.dual_exterior_power(2)._repr_()
            '2nd exterior power of the dual of the Rank-5 free module M over the Integer Ring'
            sage: M.dual_exterior_power(3)._repr_()
            '3rd exterior power of the dual of the Rank-5 free module M over the Integer Ring'
            sage: M.dual_exterior_power(4)._repr_()
            '4th exterior power of the dual of the Rank-5 free module M over the Integer Ring'
            sage: M.dual_exterior_power(5)._repr_()
            '5th exterior power of the dual of the Rank-5 free module M over the Integer Ring'
            sage: M.dual_exterior_power(21)._repr_()
            '21st exterior power of the dual of the Rank-5 free module M over the Integer Ring'
        """
        description = "{}".format(self._degree.ordinal_str())
        description += " exterior power of the dual of the {}".format(
                                                                 self._fmodule)
        return description

    def base_module(self):
        r"""
        Return the free module on which ``self`` is constructed.

        OUTPUT:

        - instance of :class:`FiniteRankFreeModule` representing the free
          module on which the exterior power is defined.

        EXAMPLES::

            sage: M = FiniteRankFreeModule(ZZ, 5, name='M')
            sage: A = M.dual_exterior_power(2)
            sage: A.base_module()
            Rank-5 free module M over the Integer Ring
            sage: A.base_module() is M
            True
        """
        return self._fmodule

    def degree(self):
        r"""
        Return the degree of ``self``.

        OUTPUT:

        - integer `p` such that ``self`` is the exterior power `\Lambda^p(M^*)`

        EXAMPLES::

            sage: M = FiniteRankFreeModule(ZZ, 5, name='M')
            sage: A = M.dual_exterior_power(2)
            sage: A.degree()
            2
            sage: M.dual_exterior_power(4).degree()
            4
        """
        return self._degree



# sage_setup: distribution = sagemath-objects
from typing import Any

class ParentWithBase:
    def __init__(self, base: Any, *args: Any, **kwds: Any) -> None: ...
    def base_extend(self, X: Any) -> Any: ...

def check_old_coerce(p: Any) -> None: ...



# sage_setup: distribution = sagemath-modules
"""
Abelian group elements

AUTHORS:

- David Joyner (2006-02); based on free_abelian_monoid_element.py, written by David Kohel.

- David Joyner (2006-05); bug fix in order

- David Joyner (2006-08); bug fix+new method in pow for negatives+fixed corresponding examples.

- David Joyner (2009-02): Fixed bug in order.

- Volker Braun (2012-11) port to new Parent base. Use tuples for immutables.

EXAMPLES:

Recall an example from abelian groups::

    sage: F = AbelianGroup(5,[4,5,5,7,8],names = list("abcde"))
    sage: (a,b,c,d,e) = F.gens()
    sage: x = a*b^2*e*d^20*e^12
    sage: x
    a*b^2*d^6*e^5
    sage: x = a^10*b^12*c^13*d^20*e^12
    sage: x
    a^2*b^2*c^3*d^6*e^4
    sage: y = a^13*b^19*c^23*d^27*e^72
    sage: y
    a*b^4*c^3*d^6
    sage: x*y
    a^3*b*c*d^5*e^4
    sage: x.list()
    [2, 2, 3, 6, 4]
"""

###########################################################################
#  Copyright (C) 2006 William Stein <wstein@gmail.com>
#  Copyright (C) 2006 David Joyner  <wdjoyner@gmail.com>
#  Copyright (C) 2012 Volker Braun  <vbraun.name@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  https://www.gnu.org/licenses/
###########################################################################

from sage.groups.abelian_gps.element_base import AbelianGroupElementBase


class AbelianGroupElement(AbelianGroupElementBase):
    """
    Elements of an
    :class:`~sage.groups.abelian_gps.abelian_group.AbelianGroup`

    INPUT:

    - ``x`` -- list/tuple/iterable of integers (the element vector)

    - ``parent`` -- the parent
      :class:`~sage.groups.abelian_gps.abelian_group.AbelianGroup`

    EXAMPLES::

        sage: F = AbelianGroup(5, [3,4,5,8,7], 'abcde')
        sage: a, b, c, d, e = F.gens()
        sage: a^2 * b^3 * a^2 * b^-4
        a*b^3
        sage: b^-11
        b
        sage: a^-11
        a
        sage: a*b in F
        True
    """
    def as_permutation(self):
        r"""
        Return the element of the permutation group ``G`` (isomorphic to the
        abelian group ``A``) associated to ``a`` in ``A``.

        EXAMPLES::

            sage: G = AbelianGroup(3, [2,3,4], names='abc'); G
            Multiplicative Abelian group isomorphic to C2 x C3 x C4
            sage: a,b,c = G.gens()
            sage: Gp = G.permutation_group(); Gp                                        # needs sage.groups
            Permutation Group with generators [(1,2), (3,4,5), (6,7,8,9)]
            sage: a.as_permutation()                                                    # needs sage.libs.gap
            (1,2)
            sage: ap = a.as_permutation(); ap                                           # needs sage.libs.gap
            (1,2)
            sage: ap in Gp                                                              # needs sage.groups sage.libs.gap
            True
        """
        from sage.libs.gap.libgap import libgap
        G = self.parent()
        A = libgap.AbelianGroup(G.gens_orders())
        phi = libgap.IsomorphismPermGroup(A)
        gens = libgap.GeneratorsOfGroup(A)
        L2 = libgap.Product([geni**Li for geni, Li in zip(gens, self.list())])
        pg = libgap.Image(phi, L2)
        return G.permutation_group()(pg)

    def word_problem(self, words):
        """
        ``G`` and ``H`` are abelian groups, ``g`` in ``G``, ``H`` is a
        subgroup of ``G`` generated by a list (words) of elements of ``G``. If
        ``self`` is in ``H``, return the expression for ``self`` as a word in
        the elements of (words).

        This function does not solve the word problem in Sage. Rather
        it pushes it over to GAP, which has optimized (non-deterministic)
        algorithms for the word problem.

        .. WARNING::

            Don't use ``E`` (or other GAP-reserved letters) as a generator
            name.

        .. TODO::

            This needs a rewrite - see stuff in the ``matrix_grp`` directory.

        EXAMPLES::

            sage: # needs sage.libs.gap
            sage: G = AbelianGroup(2, [2,3], names='xy')
            sage: x,y = G.gens()
            sage: x.word_problem([x,y])
            [[x, 1]]
            sage: y.word_problem([x,y])
            [[y, 1]]
            sage: v = (y*x).word_problem([x,y]); v  # random
            [[x, 1], [y, 1]]
            sage: prod([x^i for x,i in v]) == y*x
            True
        """
        from sage.groups.abelian_gps.abelian_group import word_problem
        return word_problem(words, self)



# sage_setup: distribution = sagemath-modules
# sage.doctest: needs sage.libs.singular
r"""
Coherent sheaves on projective schemes

EXAMPLES:

We define the Fermat cubic surface in `\PP^2` and examine its structure sheaf::

    sage: P2.<x,y,z> = ProjectiveSpace(QQ, 2)
    sage: X = P2.subscheme(x^4 + y^4 + z^4)
    sage: sh = X.structure_sheaf()
    sage: sh
    Coherent sheaf on Closed subscheme of Projective Space of dimension 2
     over Rational Field defined by: x^4 + y^4 + z^4
    sage: sh.euler_characteristic()
    -2

AUTHORS:

- Kwankyu Lee (2024-01-22): initial version

"""

from functools import cached_property
from sage.structure.sage_object import SageObject
from sage.modules.free_module import FreeModule


class CoherentSheaf(SageObject):
    r"""
    Coherent sheaf on a projective scheme.

    INPUT:

    - ``scheme`` -- the base scheme on which the sheaf is defined

    - ``module`` -- a free module or its quotient by a submodule

    - ``twist`` -- (default: 0) an integer

    This class constructs the coherent sheaf `\tilde M(n)` if `M` is the
    ``module`` and `n` is the ``twist``.
    """
    def __init__(self, scheme, module, twist=0):
        """
        Initialize ``self``.

        TESTS::

            sage: P2.<x,y,z> = ProjectiveSpace(QQ, 2)
            sage: X = P2.subscheme(x^4 + y^4 + z^4)
            sage: sheaf = X.structure_sheaf()
            sage: TestSuite(sheaf).run(skip=['_test_pickling'])
        """
        try:
            if module.is_ambient():
                module = module.quotient(module.zero_submodule())
        except AttributeError:
            pass

        assert module.cover() == module.free_cover()

        self._base_scheme = scheme
        self._module = module
        self._twist = twist

    def _cohomology(self):
        """
        Return an object that computes the cohomology.

        EXAMPLES::

            sage: P2.<x,y,z> = ProjectiveSpace(QQ, 2)
            sage: X = P2.subscheme(x^4 + y^4 + z^4)
            sage: sheaf = X.structure_sheaf()
            sage: c = sheaf._cohomology
            sage: c.H(1).dimension()
            3
            sage: c.h(1)
            3
        """
        raise NotImplementedError('_cohomology is not implemented')

    def _repr_(self):
        """
        Return the string representation of this sheaf.

        EXAMPLES::

            sage: P2.<x,y,z> = ProjectiveSpace(QQ, 2)
            sage: X = P2.subscheme(x^4 + y^4 + z^4)
            sage: s = X.structure_sheaf()
            sage: s
            Coherent sheaf on Closed subscheme of Projective Space of dimension 2
             over Rational Field defined by: x^4 + y^4 + z^4
            sage: s.twist(2)
            Twisted coherent sheaf on Closed subscheme of Projective Space of dimension 2
             over Rational Field defined by: x^4 + y^4 + z^4
        """
        sheaf = 'Twisted coherent sheaf' if self._twist else 'Coherent sheaf'
        return f'{sheaf} on {self._base_scheme}'

    def base_scheme(self):
        """
        Return the base scheme on which this sheaf is defined.

        EXAMPLES::

            sage: P2.<x,y,z> = ProjectiveSpace(QQ, 2)
            sage: X = P2.subscheme(x^4 + y^4 + z^4)
            sage: s = X.structure_sheaf()
            sage: s.base_scheme()
            Closed subscheme of Projective Space of dimension 2 over Rational Field defined by:
              x^4 + y^4 + z^4
        """
        return self._base_scheme

    def defining_twist(self):
        """
        Return the integer by which the module defining this coherent sheaf is
        twisted.

        EXAMPLES::

            sage: P2.<x,y,z> = ProjectiveSpace(QQ, 2)
            sage: X = P2.subscheme(x^4 + y^4 + z^4)
            sage: s = X.structure_sheaf(3)
            sage: s.defining_twist()
            3
        """
        return self._twist

    def defining_module(self):
        """
        Return the module defining this coherent sheaf.

        EXAMPLES::

            sage: P2.<x,y,z> = ProjectiveSpace(QQ, 2)
            sage: X = P2.subscheme(x^4 + y^4 + z^4)
            sage: s = X.structure_sheaf()
            sage: s.defining_module()
            Quotient module by Submodule of Ambient free module of rank 1 over the integral domain Quotient
             of Multivariate Polynomial Ring in x, y, z over Rational Field by the ideal (x^4 + y^4 + z^4)
             Generated by the rows of the matrix:
              []
        """
        return self._module

    def cohomology(self, r=0):
        """
        Return the dimension of the `r`-th cohomology as a vector space.

        INPUT:

        - ``r`` -- (default: 0) a non-negative integer

        EXAMPLES::

            sage: P2.<x,y,z> = ProjectiveSpace(QQ, 2)
            sage: X = P2.subscheme(x^4 + y^4 + z^4)
            sage: sheaf = X.structure_sheaf()
            sage: sheaf.cohomology(0)
            1
            sage: sheaf.cohomology(1)
            3
            sage: sheaf.cohomology(2)
            0
        """
        return self._cohomology.h(r)

    def twist(self, t=0):
        r"""
        Return the twisted sheaf `F(n)` of this sheaf `F`.

        EXAMPLES::

            sage: P2.<x,y,z> = ProjectiveSpace(QQ, 2)
            sage: X = P2.subscheme(x^4 + y^4 + z^4)
            sage: sh = X.structure_sheaf()
            sage: sh.twist(1).euler_characteristic()
            2
            sage: sh.twist(2).euler_characteristic()
            6

        TESTS::

            sage: sh.defining_twist()
            0
            sage: twisted = sh.twist(1)
            sage: twisted_twisted = twisted.twist(2)
            sage: twisted_twisted.defining_twist()
            3
        """
        return type(self)(self._base_scheme, self._module, self._twist + t)

    def euler_characteristic(self):
        """
        Return the Euler characteristic of this coherent sheaf.

        EXAMPLES::

            sage: P2.<x,y,z> = ProjectiveSpace(QQ, 2)
            sage: X = P2.subscheme(x^4 + y^4 + z^4)
            sage: sh = X.structure_sheaf()
            sage: sh.euler_characteristic()
            -2
        """
        d = self._base_scheme.dimension()
        chi = 0
        for r in range(d + 1):  # for Grothendieck's vanishing theorem
            d = self.cohomology(r)
            if r % 2:
                chi = chi - d
            else:
                chi = chi + d
        return chi


class CoherentSheaf_on_projective_space(CoherentSheaf):
    r"""
    Coherent sheaf on a projective space.

    EXAMPLES::

        sage: P2.<x,y,z> = ProjectiveSpace(QQ, 2)
        sage: P2.structure_sheaf()
        Coherent sheaf on Projective Space of dimension 2 over Rational Field
    """
    @cached_property
    def _cohomology(self):
        """
        This property keeps the cohomology object for this sheaf.

        EXAMPLES::

            sage: P2.<x,y,z> = ProjectiveSpace(QQ, 2)
            sage: P2.structure_sheaf()._cohomology
            Maruyama Method using S(0) <-- 0
        """
        from sage.schemes.projective.cohomology import MaruyamaMethod
        return MaruyamaMethod(self._module, twist=self._twist)


class CoherentSheaf_on_projective_subscheme(CoherentSheaf):
    r"""
    Coherent sheaf on a projective subscheme.

    EXAMPLES::

        sage: P2.<x,y,z> = ProjectiveSpace(QQ, 2)
        sage: X = P2.subscheme(x^4 + y^4 + z^4)
        sage: X.structure_sheaf()
        Coherent sheaf on Closed subscheme of Projective Space of dimension 2
         over Rational Field defined by: x^4 + y^4 + z^4
    """
    @cached_property
    def _cohomology(self):
        """
        This property keeps the cohomology object for this sheaf.

        EXAMPLES::

            sage: P2.<x,y,z> = ProjectiveSpace(QQ, 2)
            sage: X = P2.subscheme(x^4 + y^4 + z^4)
            sage: X.structure_sheaf()._cohomology
            Maruyama Method using S(0) <-- S(-4) <-- 0
        """
        return self.image_to_ambient_space()._cohomology

    def image_to_ambient_space(self):
        """
        Return the direct image of this sheaf to the ambient space.

        The image is with respect to the inclusion morphism from the base
        scheme into the projective space.

        INPUT:

        - ``twist`` -- (default: `0`) an integer

        EXAMPLES::

            sage: P2.<x,y,z> = ProjectiveSpace(QQ, 2)
            sage: X = P2.subscheme(x^4 + y^4 + z^4)
            sage: X.structure_sheaf().image_to_ambient_space()
            Coherent sheaf on Projective Space of dimension 2 over Rational Field
        """
        X = self._base_scheme
        A = X.ambient_space()
        S = A.coordinate_ring()

        d = self._module.degree()
        M = FreeModule(S, d)
        I = X.defining_polynomials()
        J = self._module.relations().gens()
        G = [f * M.gen(i) for i in range(d) for f in I] + [v.change_ring(S) for v in J]
        N = M.submodule(G)
        return A.coherent_sheaf(M.quotient(N), twist=self._twist)



# sage_setup: distribution = sagemath-flint
# sage.doctest: needs sage.libs.linbox
r"""
Misc matrix algorithms using FLINT
"""

from cysignals.signals cimport sig_check

from sage.arith.rational_reconstruction cimport mpq_rational_reconstruction
from sage.libs.gmp.mpq cimport *
from sage.libs.gmp.mpz cimport *
from sage.libs.flint.fmpq cimport fmpq_set_mpq, fmpq_canonicalise
from sage.libs.flint.fmpq_mat cimport fmpq_mat_entry_num, fmpq_mat_entry_den, fmpq_mat_entry
from sage.libs.flint.fmpz cimport fmpz_set_mpz, fmpz_one
from sage.rings.integer cimport Integer
from sage.rings.rational_field import QQ

from sage.matrix.matrix_integer_dense cimport Matrix_integer_dense
from sage.matrix.matrix_rational_dense cimport Matrix_rational_dense


def matrix_integer_dense_rational_reconstruction(Matrix_integer_dense A, Integer N):
    r"""
    Given a matrix over the integers and an integer modulus, do
    rational reconstruction on all entries of the matrix, viewed as
    numbers mod `N`.  This is done efficiently by assuming there is a
    large common factor dividing the denominators.

    INPUT:

    - ``A`` -- matrix
    - ``N`` -- integer

    EXAMPLES::

        sage: B = ((matrix(ZZ, 3,4, [1,2,3,-4,7,2,18,3,4,3,4,5])/3)%500).change_ring(ZZ)
        sage: from sage.matrix.misc_flint import matrix_integer_dense_rational_reconstruction
        sage: matrix_integer_dense_rational_reconstruction(B, 500)
        [ 1/3  2/3    1 -4/3]
        [ 7/3  2/3    6    1]
        [ 4/3    1  4/3  5/3]

    TESTS:

    Check that :issue:`9345` is fixed::

        sage: A = random_matrix(ZZ, 3)
        sage: matrix_integer_dense_rational_reconstruction(A, 0)
        Traceback (most recent call last):
        ...
        ZeroDivisionError: The modulus cannot be zero
    """
    if not N:
        raise ZeroDivisionError("The modulus cannot be zero")
    cdef Matrix_rational_dense R
    R = Matrix_rational_dense.__new__(Matrix_rational_dense,
                                      A.parent().change_ring(QQ), 0,0,0)

    cdef mpz_t a, bnd, other_bnd, denom, tmp
    cdef mpq_t qtmp
    cdef Integer _bnd
    cdef Py_ssize_t i, j
    cdef int do_it

    mpz_init_set_si(denom, 1)
    mpz_init(a)
    mpz_init(tmp)
    mpz_init(other_bnd)
    mpq_init(qtmp)

    _bnd = (N//2).isqrt()
    mpz_init_set(bnd, _bnd.value)
    mpz_sub(other_bnd, N.value, bnd)

    for i in range(A._nrows):
        for j in range(A._ncols):
            sig_check()
            A.get_unsafe_mpz(i, j, a)
            if mpz_cmp_ui(denom, 1) != 0:
                mpz_mul(a, a, denom)
            mpz_fdiv_r(a, a, N.value)
            do_it = 0
            if mpz_cmp(a, bnd) <= 0:
                do_it = 1
            elif mpz_cmp(a, other_bnd) >= 0:
                mpz_sub(a, a, N.value)
                do_it = 1
            if do_it:
                fmpz_set_mpz(fmpq_mat_entry_num(R._matrix, i, j), a)
                if mpz_cmp_ui(denom, 1) != 0:
                    fmpz_set_mpz(fmpq_mat_entry_den(R._matrix, i, j), denom)
                    fmpq_canonicalise(fmpq_mat_entry(R._matrix, i, j))
                else:
                    fmpz_one(fmpq_mat_entry_den(R._matrix, i, j))
            else:
                # Otherwise have to do it the hard way
                A.get_unsafe_mpz(i, j, tmp)
                mpq_rational_reconstruction(qtmp, tmp, N.value)
                mpz_lcm(denom, denom, mpq_denref(qtmp))
                fmpq_set_mpq(fmpq_mat_entry(R._matrix, i, j), qtmp)

    mpz_clear(denom)
    mpz_clear(a)
    mpz_clear(tmp)
    mpz_clear(other_bnd)
    mpz_clear(bnd)
    mpq_clear(qtmp)

    return R
